#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
AQUA basic tool for generating catalog entries based on jinja
'''

import os
import re
import sys
import argparse
import jinja2

from aqua.core.util import load_yaml, dump_yaml, get_arg
from aqua.core.configurer import ConfigPath
from aqua.core.logger import log_configure

from aqua.core.lock import SafeFileLock
from ruamel.yaml import YAML
yaml = YAML()
yaml.default_flow_style = None  # Ensure default flow style is None


def catgen_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser(description='AQUA FDB entries generator')

    parser.add_argument("-p", "--portfolio", help="Type of Data Portfolio utilized (full/reduced/minimal)")
    parser.add_argument('-c', '--config', type=str, help='yaml configuration file', required=True)
    parser.add_argument('-l', '--loglevel', type=str, help='loglevel', default='INFO')

    return parser

def get_nested(cfg, key):
    if isinstance(key, tuple):
        value = cfg
        for k in key:
            value = value.get(k) if value else None
        return value
    return cfg.get(key)

class AquaFDBGenerator:
    def __init__(self, data_portfolio, config_path, loglevel='INFO'):

        # config reading
        self.config = load_yaml(config_path)
        self.dp_version = data_portfolio

        # logging
        self.loglevel = loglevel
        self.logger = log_configure(self.loglevel, 'FDB catalog generator')

        # get the templates and config files from the AQUA installation
        self.catgendir = os.path.join(ConfigPath().configdir, 'catgen')
        self.logger.debug("Reading configuration files from %s", self.catgendir)
        self.template = self.load_jinja_template(os.path.join(self.catgendir, "catalog_entry.j2"))
        self.matching_grids = load_yaml(os.path.join(self.catgendir, "matching_grids.yaml"))


        # config options
        required_keys = [
            "author",
            "machine",
            ("repos", "data-portfolio_path"),
            ("repos", "Climate-DT-catalog_path"),
            "model",
            "resolution",
            "activity",
            "experiment",
            "expver",
            "expid",
            "data_start_date"
        ]

        # check missing parameters in config file
        missing = [k if isinstance(k, str) else ".".join(k)
                for k in required_keys if not get_nested(self.config, k)]

        if missing:
            raise ValueError(f"Missing required configuration keys: {', '.join(missing)}")

        # config options
        self.author = self.config['author']
        self.machine = self.config['machine']
        self.dp_dir_path = self.config["repos"]["data-portfolio_path"]
        self.catalog_dir_path = self.config["repos"]["Climate-DT-catalog_path"]
        self.model = self.config["model"].lower()
        #self.portfolio = self.config["portfolio"]
        self.resolution = self.config["resolution"]
        self.ocean_grid = self.config.get("ocean_grid")
        self.atm_grid = self.config.get("atm_grid")
        self.num_of_realizations = int(self.config.get("num_of_realizations", 1))
        self.description = None
        self.grid_resolutions = None


        # portfolio
        self.logger.info("Running FDB catalog generator for %s portfolio for model %s", data_portfolio, self.model)
        self.dp = load_yaml(os.path.join(self.dp_dir_path, 'portfolios', data_portfolio, 'portfolio.yaml'))
        self.grids = load_yaml(os.path.join(self.dp_dir_path, 'portfolios', data_portfolio, 'grids.yaml'))
        self.levels = load_yaml(os.path.join(self.dp_dir_path, 'definitions', 'levels.yaml'))

        self.local_grids = self.get_local_grids(self.resolution, self.grids)


    def get_local_grids(self, resolution, grids):
        """
        Get local grids based on the portfolio.

        Args:
            resolution (str): The portfolio resolution type.
            grids (dict): The grids definition.

        Returns:
            dict: Local grids for the given portfolio.
        """
        local_grids = grids["common"]
        if resolution in grids:
            self.logger.debug('Update grids for specific %s portfolio', resolution)
            local_grids.update(grids[resolution])
        else:
            self.logger.error('Cannot find grids for %s portfolio', resolution)
        print(local_grids)
        return local_grids

    def get_available_resolutions(self, local_grids, model):
        """
        Get available resolutions from local grids.

        Args:
            local_grids (dict): Local grids definition.
            model (str): The model name.

        Returns:
            list: List of available grid resolutions.
        """
        re_pattern = f"horizontal-{model.upper()}-(.+)"
        grid_resolutions = [match.group(1) for key in local_grids if (match := re.match(re_pattern, key))]
        self.logger.debug('Resolutions found are %s', grid_resolutions)
        return grid_resolutions

    @staticmethod
    def get_levelist(profile, local_grids, levels):
        """
        Get the level list from local grids.

        Args:
            profile (dict): Profile definition.
            local_grids (dict): Local grids definition.
            levels (dict): Levels definition.

        Returns:
            tuple: Levelist and levels values.
        """

        vertical = profile.get("vertical")
        if vertical is None:
            return None, None
        level_data = levels[local_grids[f"vertical-{vertical}"]]
        return level_data["levelist"], level_data["levels"]

    @staticmethod
    def get_time(frequency, levtype):
        """
        Get time string based on the frequency.

        Args:
            time_dict (dict): The time-related dictionary information

        Returns:
            str: Corresponding time string.
        """

        freq2time = {
            "hourly": {
                #'time': '"0000/to/2300/by/0100"',
                'time': '0000',
                'chunks': '6h' if levtype == 'pl' else 'D',
                'savefreq': 'h'
            },
            "6-hourly": {
                'time': '0000',
                'chunks': '6h',
                'savefreq': 'h'
            },
            "daily": {
                'time': "0000",
                'chunks': "D",
                'savefreq': "D"
            },
            "monthly": {
                'time': None,
                'chunks': 'MS',
                'savefreq': "MS"
            }
        }
        return freq2time[frequency]
    

    @staticmethod
    def get_value_from_map(value, value_map, value_type):
        """
        Get the value from the map based on the value type.
        """
        result = value_map.get(value)
        if not result:
            raise ValueError(f"Unexpected {value_type}: {value}")
        return result


    def load_jinja_template(self, template_file):
        """
        Load a Jinja2 template.

        Args:
            template_file (str): Template file name.

        Returns:
            jinja2.Template: Loaded Jinja2 template.
        """

        templateloader = jinja2.FileSystemLoader(searchpath=os.path.dirname(template_file))
        templateenv = jinja2.Environment(loader=templateloader, trim_blocks=True, lstrip_blocks=True)
        if os.path.exists(template_file):
            self.logger.debug('Loading template for %s', template_file)
            return templateenv.get_template(os.path.basename(template_file))
        
        raise FileNotFoundError(f'Cannot file template file {template_file}')

    def get_profile_content(self, profile, grid_resolution):
        """
        Generate profile content based on the given parameters.

        Args:
            profile (dict): Profile definition.
            grid_resolution (str): Resolution value.

        Returns:
            dict: Generated profile content.
        """

        grid = self.local_grids[f"horizontal-{self.model.upper()}-{grid_resolution}"]
        
        aqua_grid = self.matching_grids[grid]
        levelist, levels_values = self.get_levelist(profile, self.local_grids, self.levels)

        levtype_str = (
            'atm2d' if profile["levtype"] == 'sfc' else
            'atm3d' if profile["levtype"] == 'pl' else
            'oce2d' if profile["levtype"] == 'o2d' else
            'oce3d' if profile["levtype"] == 'o3d' and 'full' in profile['vertical'] else
            'oce3d-half' if profile["levtype"] == 'o3d' and 'half' in profile['vertical'] else
            'sol4' if profile["levtype"] == 'sol' and profile['vertical'] == 'IFS-sol4' or profile['vertical'] == 'ICON-sol4' else
            'sol5' if profile["levtype"] == 'sol' and profile['vertical'] == 'IFS-sol5' or profile['vertical'] == 'ICON-sol5' else
            profile["levtype"] 
        )

        if not self.ocean_grid:
            self.ocean_grid = self.matching_grids['ocean_grid'][self.model][self.resolution]
            if self.ocean_grid is None:
                raise ValueError(f"No ocean grid available for: {self.model} {self.resolution}")

        if not self.atm_grid:
            self.atm_grid = self.matching_grids['atm_grid'][self.model][self.resolution]
            if self.atm_grid is None:
                raise ValueError(f"No atmospheric grid available for: {self.model} {self.resolution}")
                
        grid_mappings = self.matching_grids['grid_mappings']
        levtype = profile["levtype"]

        if levtype in grid_mappings:
            grid_str = grid_mappings[levtype].get(
                self.model, grid_mappings[levtype].get('default')).format(ocean_grid=self.ocean_grid, aqua_grid=aqua_grid)
        else:
            grid_str = grid_mappings['default'].format(aqua_grid=aqua_grid)
 
        source = f"{profile['frequency']}-{aqua_grid}-{levtype_str}"
        self.logger.info('Source: %s', source)

        self.logger.debug('levtype: %s, levels: %s, grid: %s', levtype, levelist, grid_str)

        time_dict = self.get_time(frequency=profile["frequency"], levtype=profile['levtype'])
        self.logger.debug('Time dict: %s', time_dict)
        self.logger.debug('Number of realizations %s', self.num_of_realizations)

        self.description = (
            self.config.get("description")
            or f'"{self.model} {self.config["exp"]} {self.config["data_start_date"][:4]}, '
            f'grids: {self.atm_grid} {self.ocean_grid}"' )
        
        # Set the stream based on the frequency
        stream = 'clmn' if profile['frequency'] == 'monthly' else 'clte'

        kwargs = {
            "dp_version": self.dp_version,
            "resolution": grid_resolution,
            "grid": grid_str,
            "source": source,
            "levelist": levelist,
            "num_of_realizations": self.num_of_realizations,
            "levels": levels_values,
            "levtype": profile["levtype"],
            "stream": stream,
            "variables": profile["variables"],
            "param": profile["variables"][0],
            "time": time_dict['time'],
            "chunks": time_dict['chunks'],
            "savefreq": time_dict['savefreq'], 
            "description": self.description
        }
        return kwargs

    def create_catalog_entry(self, all_content):
        """
        Create catalog entry file and update main YAML.

        Args:
            all_content (dict): Dictionary of all generated content strings.
        """
        output_dir = os.path.join(self.catalog_dir_path, 'catalogs',
                                  self.config['catalog_dir'], 'catalog', self.model.upper())
        os.makedirs(output_dir, exist_ok=True)
        output_filename = f"{self.config['exp']}.yaml"
        output_path = os.path.join(output_dir, output_filename)

        # Remove the existing file if it exists
        if os.path.exists(output_path):
            os.remove(output_path)

        # Write the catalog entry file
        dump_yaml(output_path, all_content)
        self.logger.info("File %s has been created in %s", output_filename, output_dir)

        # Update main.yaml
        main_yaml_path = os.path.join(output_dir, 'main.yaml')
        with SafeFileLock(main_yaml_path + '.lock', loglevel=self.loglevel):
            if not os.path.exists(main_yaml_path):
                main_yaml = {'sources': {}}
            else:
                main_yaml = load_yaml(main_yaml_path)

            resolution_map = {
                'production': 'HR',
                'develop': 'SR',
                'lowres': 'LR',
                'intermediate': 'MR'
            }
            
            resolution_id = self.get_value_from_map(self.config['resolution'], resolution_map, 'resolution')

            forcing_map = {
                'hist': 'historical',
                'cont': 'control',
                'SSP3-7.0': 'ssp370',
                'Tplus2.0K': 'tplus2K'
            }

            forcing = self.config.get('forcing')
            if not forcing:
                experiment = self.config['experiment']
                forcing = forcing_map.get(experiment, re.sub(r'[^a-z0-9]', '', experiment.lower()))
            
            main_yaml['sources'][self.config['exp']] = {
                'description': self.description,
                'metadata': {
                    'author': self.author,
                    'maintainer': self.config.get('maintainer') or 'not specified',
                    'machine': self.machine,
                    'expid': self.config['expid'],                
                    'resolution_atm': self.atm_grid,
                    'resolution_oce': self.ocean_grid,
                    'forcing': forcing,
                    'start': self.config['data_start_date'][:4], #year only
                    'dashboard': {
                        'menu': self.config.get('menu') or self.config['exp'],
                        'resolution_id': resolution_id,
                        'note': self.config.get('note')
                    }
                },
                'driver': 'yaml_file_cat',
                'args': {
                    'path': f"{{{{CATALOG_DIR}}}}/{self.config['exp']}.yaml"
                }
            }
            dump_yaml(main_yaml_path, main_yaml)
            self.logger.info("%s entry in 'main.yaml' has been updated in %s", self.config['exp'], output_dir)

        # Update catalog.yaml if a new model is added
        catalog_yaml_path = os.path.join(self.catalog_dir_path, 'catalogs',  self.config['catalog_dir'], 'catalog.yaml')
        with SafeFileLock(catalog_yaml_path + '.lock', loglevel=self.loglevel):
            catalog_yaml = load_yaml(catalog_yaml_path)

            if catalog_yaml.get('sources') is None:
                catalog_yaml['sources'] = {}

            if self.model not in catalog_yaml.get('sources', {}):  
                catalog_yaml.setdefault('sources', {}) 
                catalog_yaml['sources'][self.model.upper()] = {
                    'description': f"{self.model.upper()} model",
                    'driver': 'yaml_file_cat',
                    'args': {
                        'path': f"{{{{CATALOG_DIR}}}}/catalog/{self.model.upper()}/main.yaml"
                    }
                }
                dump_yaml(catalog_yaml_path, catalog_yaml)
                self.logger.info("%s entry in 'catalog.yaml' has been created at %s", self.model, catalog_yaml_path)


    def generate_catalog(self):
        """
        Generate the entire catalog by processing profiles and resolutions.
        """
        all_content = {'sources': {}}

        # Retrieve available resolutions for the current model
        self.grid_resolutions = self.get_available_resolutions(self.local_grids, self.model)
        
        if not self.grid_resolutions:
            self.logger.error('No resolutions found, generating an empty file!')
            return

        for profile in self.dp[self.model]:

            # Filter out omitted resolutions, if any
            current_resolutions = [
                res for res in self.grid_resolutions 
                if 'omit-resolutions' not in profile or res not in profile['omit-resolutions']
            ]

            for grid_resolution in current_resolutions:
        
                content = self.get_profile_content(profile, grid_resolution)
                combined = {**self.config, **content}
                source_name = combined.get('source')

                if source_name in all_content['sources']:
                    self.logger.debug('Source %s already exists, updating variables.', source_name)

                    source_content = all_content['sources'][source_name]
                    source_content['metadata']['variables'].extend(combined['variables'])

                    all_content['sources'][source_name] = source_content
                    self.logger.info('Added variables %s to source %s.', combined['variables'], source_name)
                else:
                    self.logger.debug('Creating catalog entry for %s.', source_name)
                    # Convert lists to inline format before rendering for better readability
                    for key in ['levels', 'variables']:
                        if key in combined and combined[key] is not None:
                            combined[key] = list(combined[key])
                    try:
                        rendered_content = yaml.load(self.template.render(combined))
                        all_content['sources'][source_name] = rendered_content[source_name]
                    except Exception as e:
                        self.logger.error('Error rendering template for source %s: %s', source_name, str(e))

        # Create final catalog entry
        self.create_catalog_entry(all_content)

def catgen_execute(args):
    """Useful wrapper for the FDB catalog generator class"""

    dp_version = get_arg(args, 'portfolio', 'full')
    config_file = get_arg(args, 'config', 'config.yaml')
    loglevel = get_arg(args, 'loglevel', 'INFO')

    generator = AquaFDBGenerator(dp_version, config_file, loglevel)
    generator.generate_catalog()


if __name__ == '__main__':

    args = catgen_parser().parse_args(sys.argv[1:])
    catgen_execute(args)




import intake  # Import this first to avoid circular imports during discovery.
# from intake.container import register_container

from .intake_gsv import GSVSource

try:
    intake.registry.drivers.register_driver('gsv', GSVSource)
except ValueError:
    pass

# register_container('gsv', GSVSource)



"""Test checking if all catalog entries can be read"""

import pytest
import xarray
from aqua import Reader
from aqua.core.reader import show_catalog_content as catalog
from conftest import LOGLEVEL


@pytest.fixture(params=[(model, exp, source)
                        for model in catalog(catalog_name="ci", verbose=False)['ci']
                        for exp in catalog(catalog_name="ci", verbose=False)['ci'][model]
                        for source in catalog(catalog_name="ci", verbose=False)['ci'][model][exp]])
def reader(request):
    """Reader instance fixture"""
    model, exp, source = request.param
    if source == 'intake-esm-test': # temporary skip of intake esm sources
        pytest.skip("Skipping intake-esm-test for now, not supported for now")
    myread = Reader(catalog='ci', model=model, exp=exp, source=source, areas=False,
                    fix=False, loglevel=LOGLEVEL)
    data = myread.retrieve()
    return myread, data

@pytest.mark.gsv
def test_catalog_gsv():
    """
    Checking that both reader and Dataset are retrived in reasonable shape
    """
    sources = ['fdb', 'fdb-levels', 'fdb-nolevels']

    for source in sources:
        reader_gsv = Reader(model='IFS', exp='test-fdb', source=source,
                            loglevel=LOGLEVEL)
        data = reader_gsv.retrieve()

        assert isinstance(reader_gsv, Reader)
        assert isinstance(data, xarray.Dataset)

@pytest.fixture(params=[(model, exp, source)
                        for model in catalog(catalog_name="ci", verbose=False)['ci']
                        for exp in catalog(catalog_name="ci")['ci'][model]
                        for source in catalog(catalog_name="ci")['ci'][model][exp]])
def reader_regrid(request):
    """Reader instance fixture"""
    model, exp, source = request.param
    print([model, exp, source])
    myread = Reader(catalog='ci', model=model, exp=exp, source=source, areas=True, regrid='r200',
                    loglevel=LOGLEVEL, rebuild=False)
    data = myread.retrieve()

    return myread, data


@pytest.mark.slow
def test_catalog(reader):
    """
    Checking that both reader and Dataset are retrived in reasonable shape
    """
    aaa, bbb = reader
    assert isinstance(aaa, Reader)
    assert isinstance(bbb, xarray.Dataset)


@pytest.mark.sbatch
def test_catalog_reader(reader_regrid):
    """
    Checking that data can be regridded
    """
    read, data = reader_regrid
    vvv = list(data.data_vars)[-1]
    select = data[vvv].isel(time=0)
    rgd = read.regrid(select)
    assert len(rgd.lon) == 180
    assert len(rgd.lat) == 90

# @pytest.mark.aqua
# @pytest.mark.parametrize(
#     "catalog, model, exp, source, expected_output",
#     [
#         # Test case 1: Source is specified and exists in the catalog
#         (
#             {"model1": {"exp1": {"source1": "data1", "source2": "data2"}}},
#             "model1",
#             "exp1",
#             "source1",
#             "source1"
#         ),
#         # Test case 2: Source is specified but does not exist,
#         # default source exists
#         (
#             {"model1": {"exp1": {"default": "default_data", "source2": "data2"}}},
#             "model1",
#             "exp1",
#             "source1",
#             "default"
#         ),
#         # Test case 3: Source is specified but does not exist,
#         # default source does not exist
#         (
#             {"model1": {"exp1": {"source2": "data2"}}},
#             "model1",
#             "exp1",
#             "source1",
#             pytest.raises(KeyError)
#         ),
#         # Test case 4: Source is not specified, choose the first source
#         (
#             {"model1": {"exp1": {"source1": "data1", "source2": "data2"}}},
#             "model1",
#             "exp1",
#             None,
#             "source1"
#         ),
#         # Test case 5: Source is not specified, no sources available
#         (
#             {"model1": {"exp1": {}}},
#             "model1",
#             "exp1",
#             None,
#             pytest.raises(KeyError)
#         ),
#         # Test case 6: Source is not specified, no sources available,
#         # but a default source exists
#         (
#             {"model1": {"exp1": {"default": "default_data"}}},
#             "model1",
#             "exp1",
#             None,
#             "default"
#         )
#     ]
# )
# def test_check_catalog_source(catalog, model, exp, source, expected_output):
#     if isinstance(expected_output, str):
#         assert check_catalog_source(catalog, model, exp, source) == expected_output
#     else:
#         with expected_output:
#             check_catalog_source(catalog, model, exp, source)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
AQUA files operations mixin
'''

import os
import sys
import shutil

from aqua.core.lock import SafeFileLock
from aqua.core.util import load_yaml, dump_yaml, load_multi_yaml


class FilesMixin:
    """Mixin for AQUA file operations"""

    def fixes_add(self, args):
        """Add a fix file

        Args:
            args (argparse.Namespace): arguments from the command line
        """
        compatible = self._check_file(kind='fixes', file=args.file)
        if compatible:
            self._file_add(kind='fixes', file=args.file, link=args.editable)

    def grids_add(self, args):
        """Add a grid file

        Args:
            args (argparse.Namespace): arguments from the command line
        """
        compatible = self._check_file(kind='grids', file=args.file)
        if compatible:
            self._file_add(kind='grids', file=args.file, link=args.editable)

    def grids_set(self, args):
        """
        Set the grids (and concurrently the weights and areas) paths in the config-aqua.yaml
        This will override the grids paths defined in the individual catalogs

        Args:
            args (argparse.Namespace): arguments from the command line
        """
        self._check()
        grids_path = args.path + '/grids'
        areas_path = args.path + '/areas'
        weights_path = args.path + '/weights'

        self.logger.info('Setting grids path to %s, weights path to %s and areas path to %s',
                         grids_path, weights_path, areas_path)

        # Check if the paths exist and if not create them
        for path in [grids_path, areas_path, weights_path]:
            if not os.path.exists(path):
                self.logger.info('Creating path %s', path)
                os.makedirs(path, exist_ok=True)

        filename = os.path.join(self.configpath, self.configfile)
        with SafeFileLock(filename + '.lock', loglevel=self.loglevel):
            cfg = load_yaml(filename)
            path_dict = {
                'paths': {
                    'grids': grids_path,
                    'areas': areas_path,
                    'weights': weights_path
                }
            }

            # If the paths already exist, we just update them
            if 'paths' in cfg:
                self.logger.info('Updating existing paths in %s', self.configfile)
                cfg['paths'].update(path_dict['paths'])
            else:
                self.logger.info('Adding new paths to %s', self.configfile)
                cfg['paths'] = path_dict['paths']

            dump_yaml(filename, cfg)

    def _file_add(self, kind, file, link=False):
        """Add a personalized file to the fixes/grids folder

        Args:
            kind (str): the kind of file to be added, either 'fixes' or 'grids'
            file (str): the file to be added
            link (bool): whether to add the file as a link or not
        """

        file = os.path.abspath(file)
        self._check()
        basefile = os.path.basename(file)
        pathfile = f'{self.configpath}/{kind}/{basefile}'
        if not os.path.exists(pathfile):
            if link:
                self.logger.info('Linking %s to %s', file, pathfile)
                os.symlink(file, pathfile)
            else:
                self.logger.info('Installing %s to %s', file, pathfile)
                shutil.copy(file, pathfile)
        else:
            self.logger.error('%s for file %s already installed, or a file with the same name exists', kind, file)
            sys.exit(1)

    def remove_file(self, args):
        """Add a personalized file to the fixes/grids folder

        Args:
            kind (str): the kind of file to be added, either 'fixes' or 'grids'
            file (str): the file to be added
        """

        self._check()
        kind = args.command
        file = os.path.basename(args.file)
        pathfile = f'{self.configpath}/{kind}/{file}'
        if os.path.exists(pathfile):
            self.logger.info('Removing %s', pathfile)
            if os.path.islink(pathfile):
                os.unlink(pathfile)
            else:
                os.remove(pathfile)
        else:
            self.logger.error('%s file %s is not installed in AQUA, cannot remove it',
                              kind, file)
            sys.exit(1)

    def _check_file(self, kind, file=None):
        """
        Check if a new file can be merged with AQUA load_multi_yaml()
        It works also without a new file to check that the existing files are compatible

        Args:
            kind (str): the kind of file to be added, either 'fixes' or 'grids'
            file (str): the file to be added
        """
        if kind not in ['fixes', 'grids']:
            raise ValueError('Kind must be either fixes or grids')

        self._check()
        try:
            _ = load_multi_yaml(folder_path=f'{self.configpath}/{kind}',
                            filenames=[file]) if file is not None else load_multi_yaml(folder_path=f'{self.configpath}/{kind}')

            if file is not None:
                self.logger.debug('File %s is compatible with the existing files in %s', file, kind)

            return True
        except Exception as e:
            if file is not None:
                if not os.path.exists(file):
                    self.logger.error('%s is not a valid file!', file)
                else:
                    self.logger.error("It is not possible to add the file %s to the %s folder", file, kind)
            else:
                self.logger.error("Existing files in the %s folder are not compatible", kind)
            self.logger.error(e)
            return False



"""Module for custom exceptions used in AQUA"""


class NoDataError(Exception):
    """Exception raised when there is no data available"""

    def __init__(self, message="No data available"):
        self.message = message
        super().__init__(self.message)


class NoEcCodesShortNameError(Exception):
    """Exception raised when there is no ecCodes shortName available"""

    def __init__(self, message="No ecCodes shortName available"):
        self.message = message
        super().__init__(self.message)


class NotEnoughDataError(Exception):
    """Exception raised when there is not enough data available"""

    def __init__(self, message="Not enough data available"):
        self.message = message
        super().__init__(self.message)


class NoObservationError(Exception):
    """Exception raised when there is no observation available"""

    def __init__(self, message="No observation available"):
        self.message = message
        super().__init__(self.message)

class NoRegridError(Exception):
    """Exception raised when no regrid is available"""

    def __init__(self, message="No regrid available"):
        self.message = message
        super().__init__(self.message)



"""Utility functions for console module."""

import sys

def query_yes_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    Args:
        question (str): the question to be asked to the user
        default (str): the default answer if the user just hits <Enter>.
                        It must be "yes" (the default), "no" or None (meaning
                        an answer is required of the user).

    Returns:
        bool: True for yes, False for no
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError(f"invalid default answer: {default}")

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        if choice in valid:
            return valid[choice]
        print("Please respond with 'yes' or 'no' (or 'y' or 'n').")


import pytest
from aqua import Reader
from conftest import LOGLEVEL

@pytest.mark.aqua
class TestKwargs:
    """Small class to test kwargs"""

    def test_kwargs_default(self):
        reader = Reader(model="FESOM", exp="test-pi", source="kwargs-2d", loglevel=LOGLEVEL)
        data = reader.retrieve()
        assert list(data.data_vars) == ['tos']
        assert data.time.dt.year == 1985

    @pytest.mark.aqua
    def test_kwargs_one_option(self):
        reader = Reader(model="FESOM", exp="test-pi", source="kwargs-2d", year=1986, loglevel=LOGLEVEL)
        data = reader.retrieve()
        assert list(data.data_vars) == ['tos']
        assert data.time.dt.year == 1986

    @pytest.mark.aqua
    def test_kwargs_two_options(self):
        reader = Reader(model="FESOM", exp="test-pi", source="kwargs-2d", year=1986, variable='a_ice', fix=False, loglevel=LOGLEVEL)
        data = reader.retrieve()
        assert list(data.data_vars) == ['a_ice']
        assert data.time.dt.year == 1986

    @pytest.mark.aqua
    def test_wrong_kwargs(self):
        # With a wrong kwarg, the reader will remove it and log a warning
        reader = Reader(model="FESOM", exp="test-pi", source="kwargs-2d", wrong_kwarg='pippo', loglevel=LOGLEVEL)
        data = reader.retrieve()
        assert list(data.data_vars) == ['tos']
        assert data.time.dt.year == 1985



"""Utility module for DROP"""

import os
import shutil
from glob import glob
# from aqua.core.util import dump_yaml, load_yaml
# from aqua.core.util import ConfigPath
# from aqua.core.logger import log_configure


# def opa_catalog_entry(datadir, model, exp, source, catalog=None,
#                       fixer_name=False, frequency='monthly',
#                       loglevel='WARNING'):
#     """
#     Create an entry in the AQUA catalog based on the presence of output from OPA in datadir
#     to be used by DROP in both source and regrid yaml

#     Args:
#         datadir (str): path to the directory containing the data
#         catalog (str): name of the catalog, None for getting it automatically
#         model (str): name of the model
#         exp (str): name of the experiment
#         source (str): name of the origin source
#         fixer_name (str): fix to be used when reading the opa. Default is False
#         frequency (str, opt): frequency of the data, default is 'monthly'
#         loglevel (str, opt): logging level, default is 'WARNING'

#     Returns:
#         entry_name (str): name of the entry created in the catalog

#     Raises:
#         KeyError: if the origin source is not found in the catalog
#     """
#     logger = log_configure(log_level=loglevel, log_name='opa_catalog_entry')

#     entry_name = 'opa'
#     entry_name += '_%s' % source
#     logger.info('Creating catalog entry %s %s %s', model, exp, entry_name)

#     # load the catalog experiment file
#     Configurer = ConfigPath()
#     configdir = Configurer.configdir
#     if catalog is None:
#         catalog = Configurer.catalog

#     # find the catalog of my experiment
#     catalogfile = os.path.join(configdir, 'catalogs', catalog,
#                                'catalog', model, exp + '.yaml')

#     # load, add the block and close
#     cat_file = load_yaml(catalogfile)

#     # NOTE: commented because the correct grid is lon-lat when generated by OPA
#     #       see issue #691
#     # # read the grid info from the origin source
#     # if source in cat_file['sources']:
#     #     try:
#     #         grid = cat_file['sources'][source]['metadata']['source_grid_name']
#     #     except KeyError:
#     #         logger.error('Cannot find source grid name in catalog, assuming lon-lat')
#     #         grid = 'lon-lat'
#     # else:
#     #     raise KeyError('Cannot find source %s in catalog' % source)

#     # define the block to be uploaded into the catalog
#     description = 'OPA output from %s as origin source' % source

#     block_cat = {
#         'driver': 'netcdf',
#         'args': {
#             'urlpath': os.path.join(datadir, f'*{frequency}_mean.nc'),
#             'chunks': {},
#             'xarray_kwargs': {
#                 'decode_times': True,
#                 'combine': 'by_coords'
#             }
#         },
#         'description': description,
#         'metadata': {
#             'source_grid_name': 'lon-lat',
#             'fixer_name': fixer_name
#         }
#     }

#     # add the block to the catalog
#     cat_file['sources'][entry_name] = block_cat
#     dump_yaml(outfile=catalogfile, cfg=cat_file)

#     return entry_name


def list_drop_files_complete(path):
    """
    List DROP-generated files in the specified path based on the given parameters.

    Args:
        path (str): The base path where the DROP-generated files are located.

    Returns:
        A dictionary containing the netcdf files for a each DROP variable
    """

    yearly_dict = {}
    monthly_dict = {}
    searchpath = "*"
    yearly_dict['files'] = sorted(glob(os.path.join(path, searchpath) + '_????.nc'))
    monthly_dict['files'] = sorted(glob(os.path.join(path, searchpath) + '_??????.nc'))

    # path = os.path.join(path, model, exp, reso, freq)
    return yearly_dict, monthly_dict


# NOT USED AT THE MOMENT
# def list_drop_files_vars(path):
#     """
#     List DROP files in the specified path based on the given parameters.

#     Args:
#         path (str): The base path where the DROP files are located.


#     """

#     # path = os.path.join(path, model, exp, reso, freq)
#     searchpath = os.path.join(path, '*.nc')
#     variables = set([os.path.basename(complete).rpartition('_')[0] for complete in glob(searchpath)])
#     yearly_dict = {}
#     monthly_dict = {}
#     for var in variables:
#         yearly_dict[var] = sorted(glob(os.path.join(path, var) + '_????.nc'))
#         monthly_dict[var] = sorted(glob(os.path.join(path, var) + '_??????.nc'))

#     return yearly_dict, monthly_dict

def move_tmp_files(tmp_directory, output_directory):
    """
    Move temporary NetCDF files from the tmp directory to the output directory,
    changing their name by removing "_tmp" suffix.
    """
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    for tmp_file in os.listdir(tmp_directory):
        if tmp_file.endswith(".nc"):
            if "_tmp" in tmp_file:
                new_file_name = tmp_file.replace("_tmp", "")
            else:
                new_file_name = tmp_file
            tmp_file_path = os.path.join(tmp_directory, tmp_file)
            new_file_path = os.path.join(output_directory, new_file_name)
            shutil.move(tmp_file_path, new_file_path)



"""
DROP (Data Reduction OPerator) class

This class provides comprehensive data processing capabilities for climate datasets,
including regridding, temporal averaging, regional extraction, and archiving.
It handles multiple file formats and uses Dask for parallel processing of large datasets.

Main features:
- Regridding to arbitrary resolutions
- Temporal resampling with various statistics (mean, std, max, min)
- Regional data extraction
- Automatic catalog entry generation
- Parallel processing with Dask
- Memory-efficient chunked processing
"""
import os
from time import time
import subprocess
import glob
import shutil
import dask
import xarray as xr
import numpy as np
import pandas as pd

from dask.distributed import Client, LocalCluster, progress, performance_report
from dask.diagnostics import ProgressBar
from dask.distributed.diagnostics import MemorySampler

from aqua.core.lock import SafeFileLock
from aqua.core.logger import log_configure, log_history
from aqua.core.reader import Reader
from aqua.core.util.io_util import create_folder, file_is_complete
from aqua.core.util import dump_yaml, load_yaml
from aqua.core.configurer import ConfigPath
from aqua.core.util import create_zarr_reference, replace_intake_vars
from aqua.core.util.string import generate_random_string
from .drop_util import move_tmp_files, list_drop_files_complete
from .catalog_entry_builder import CatalogEntryBuilder


TIME_ENCODING = {
    'units': 'days since 1850-01-01 00:00:00',
    'calendar': 'standard',
    'dtype': 'float64'}

VAR_ENCODING = {
    'dtype': 'float64',
    'zlib': True,
    'complevel': 1,
    '_FillValue': np.nan
}


class Drop():
    """
    Class to generate DROP outputs at required frequency/resolution
    """

    @property
    def dask(self):
        """Check if dask is needed"""
        return self.nproc > 1

    def __init__(self,
                 catalog=None, model=None, exp=None, source=None,
                 var=None, configdir=None,
                 resolution=None, frequency=None, fix=True,
                 startdate=None, enddate=None,
                 outdir=None, tmpdir=None, nproc=1,
                 loglevel=None,
                 region=None, drop=False,
                 overwrite=False, definitive=False,
                 performance_reporting=False,
                 rebuild=False,
                 exclude_incomplete=False,
                 stat="mean",
                 compact="xarray",
                 cdo_options=["-f", "nc4", "-z", "zip_1"],
                 **kwargs):
        """
        Initialize the DROP class

        Args:
            catalog (string):        The catalog you want to read. If None, guessed by the reader.
            model (string):          The model name from the catalog
            exp (string):            The experiment name from the catalog
            source (string):         The sourceid name from the catalog
            var (str, list):         Variable(s) to be processed and archived.
            resolution (string):     The target resolution for the DROP output. If None,
                                     no regridding is performed.
            frequency (string,opt):  The target frequency for averaging the
                                     DROP output, if no frequency is specified,
                                     no time average is performed
            fix (bool, opt):         True to fix the data, default is True
            startdate (string,opt): Start date for the data to be processed,
                                     format YYYYMMDD, default is None
            enddate (string,opt):   End date for the data to be processed,
                                     format YYYYMMDD, default is None
            outdir (string):         Where the DROP output is stored.
            tmpdir (string):         Where to store temporary files,
                                     default is None.
                                     Necessary for dask.distributed
            configdir (string):      Configuration directory where the catalog
                                     are found
            nproc (int, opt):        Number of processors to use. default is 1
            loglevel (string, opt):  Logging level
            region (dict, opt):      Region to be processed, default is None,
                                     meaning 'global'.
                                     Requires 'name' (str), 'lon' (list) and 'lat' (list)
            drop (bool, opt):        Drop the missing values in the region selection.
            overwrite (bool, opt):   True to overwrite existing files, default is False
            definitive (bool, opt):  True to create the output file,
                                     False to just explore the reader
                                     operations, default is False
            performance_reporting (bool, opt): True to save an html report of the
                                               dask usage, default is False. This will run a single month
                                               to collect the performance data.
            exclude_incomplete (bool,opt)   : True to remove incomplete chunk
                                            when averaging, default is false.
            rebuild (bool, opt):     Rebuild the weights when calling the reader
            stat (string, opt):      Statistic to compute. Can be 'mean', 'std', 'max', 'min'.
            compact (string, opt):   Compact the data into yearly files using xarray or cdo.
                                     If set to None, no compacting is performed. Default is "xarray"
            cdo_options (list, opt): List of options to be passed to cdo, default is ["-f", "nc4", "-z", "zip_1"]
            **kwargs:                kwargs to be sent to the Reader, as 'zoom' or 'realization'
        """

        # Check mandatory parameters
        self.catalog = self._require_param(catalog, "catalog")
        self.model = self._require_param(model, "model")
        self.exp = self._require_param(exp, "experiment")
        self.source = self._require_param(source, "source")
        self.var = self._require_param(var, "variable string or list.")

        # General settings
        self.logger = log_configure(loglevel, 'DROP')
        self.loglevel = loglevel

        # save parameters
        self.resolution = resolution
        self.frequency = frequency
        self.overwrite = overwrite
        self.exclude_incomplete = exclude_incomplete
        self.definitive = definitive
        self.nproc = int(nproc)
        self.rebuild = rebuild
        self.kwargs = kwargs
        self.fix = fix

        # configure start date and end date
        self.startdate = startdate
        self.enddate = enddate
        self._validate_date(startdate, enddate)

        # configure statistics
        self.stat = stat
        if self.stat not in ['mean', 'std', 'max', 'min']:
            raise KeyError('Please specify a valid statistic: mean, std, max or min.')

        # configure regional selection
        self._configure_region(region, drop)

        # print some info about the settings
        self._issue_info_warning()

        # define the tmpdir
        if tmpdir is None:
            self.logger.warning('No tmpdir specifield, will use outdir')
            self.tmpdir = os.path.join(outdir, 'tmp')
        else:
            self.tmpdir = tmpdir
        self.tmpdir = os.path.join(self.tmpdir, f'DROP_{generate_random_string(10)}')

        # set up compacting method for concatenation
        self.compact = compact
        if self.compact not in ['xarray', 'cdo', None]:
            raise KeyError('Please specify a valid compact method: xarray, cdo or None.')

        self.cdo_options = cdo_options
        if not isinstance(self.cdo_options, list):
            raise TypeError('cdo_options must be a list.')

        # configure the configdir
        configpath = ConfigPath(configdir=configdir)
        self.configdir = configpath.configdir

        # get default grids
        _, grids_path = configpath.get_reader_filenames()
        self.default_grids = load_yaml(os.path.join(grids_path, 'default.yaml'))

        # option for encoding, defined once for all
        self.time_encoding = TIME_ENCODING
        self.var_encoding = VAR_ENCODING

        # add the performance report
        self.performance_reporting = performance_reporting

        # Create output folders
        if outdir is None:
            raise KeyError('Please specify outdir.')

        self.catbuilder = CatalogEntryBuilder(
            catalog=self.catalog, model=self.model,
            exp=self.exp, resolution=self.resolution, frequency=self.frequency,
            region=self.region_name, stat=self.stat, loglevel=self.loglevel, **self.kwargs
        )
        # Create output path builder from the catalog entry builder
        self.outbuilder = self.catbuilder.opt

        self.basedir = outdir
        self.outdir = os.path.join(self.basedir, self.outbuilder.build_directory())

        create_folder(self.outdir, loglevel=self.loglevel)
        create_folder(self.tmpdir, loglevel=self.loglevel)

        # Initialize variables used by methods
        self.data = None
        self.cluster = None
        self.client = None
        self.reader = None

        # for data reading from FDB
        self.last_record = None
        self.check = False

    @staticmethod
    def _require_param(param, name, msg=None):
        if param is not None:
            return param
        raise KeyError(msg or f"Please specify {name}.")

    @staticmethod
    def _validate_date(startdate, enddate):
        """Validate date format for startdate and enddate"""
        
        if startdate is not None:
            try:
                pd.to_datetime(startdate)
            except (ValueError, TypeError):
                raise ValueError('startdate must be a valid date string (YYYY-MM-DD or YYYYMMDD)')
        
        if enddate is not None:
            try:
                pd.to_datetime(enddate)
            except (ValueError, TypeError):
                raise ValueError('enddate must be a valid date string (YYYY-MM-DD or YYYYMMDD)')

    def _issue_info_warning(self):
        """
        Print information about the DROP settings
        """

        if self.startdate is not None or self.enddate is not None:
            self.logger.info('startdate is %s, enddate is %s', self.startdate, self.enddate)
            self.logger.info('startdate or enddate are set, please be sure to process one experiment at the time.')

        if not self.frequency:
            self.logger.info('Frequency not specified, no time averaging will be performed.')
        else:
            self.logger.info('Frequency: %s', self.frequency)

        if self.overwrite:
            self.logger.warning('File will be overwritten if already existing.')

        if self.exclude_incomplete:
            self.logger.info('Exclude incomplete for time averaging activated!')

        if not self.definitive:
            self.logger.warning('IMPORTANT: no file will be created, this is a dry run')

        if self.dask:
            self.logger.info('Running dask.distributed with %s workers', self.nproc)

        if self.rebuild:
            self.logger.info('rebuild=True! DROP will rebuild weights and areas!')

        self.logger.info('Variable(s) to be processed: %s', self.var)
        self.logger.info('Fixing data: %s', self.fix)
        self.logger.info('Resolution: %s', self.resolution)
        self.logger.info('Statistic to be computed: %s', self.stat)
        self.logger.info('Domain selection: %s', self.region_name)

    def _configure_region(self, region, drop):
        """ Configure the region for regional selection, and the drop option"""

        if region is not None:
            self.region = region
            if self.region['name'] is None:
                raise KeyError('Please specify name in region.')
            if self.region['lon'] is None and self.region['lat'] is None:
                raise KeyError(f"Please specify at least one between lat and lon for {region['name']}.")
            self.region_name = self.region['name']
            self.logger.info('Regional selection active! region: %s, lon: %s and lat: %s...',
                             self.region['name'], self.region['lon'], self.region['lat'])
        else:
            self.region = None
            self.region_name = None
        self.drop = drop

    def retrieve(self):
        """
        Retrieve data from the catalog
        """

        # Initialize the reader
        self.reader = Reader(model=self.model, exp=self.exp,
                             source=self.source,
                             regrid=self.resolution,
                             catalog=self.catalog,
                             loglevel=self.loglevel,
                             rebuild=self.rebuild,
                             startdate=self.startdate,
                             enddate=self.enddate,
                             fix=self.fix, **self.kwargs)

        self.logger.info('Accessing catalog for %s-%s-%s...',
                         self.model, self.exp, self.source)

        if self.catalog is None:
            self.logger.info('Assuming catalog from the reader so that is %s', self.reader.catalog)
            self.catalog = self.reader.catalog

        self.logger.info('Retrieving data...')
        self.data = self.reader.retrieve(var=self.var)

        self.logger.debug(self.data)

    def drop_generator(self):
        """
        Generate DROP output
        """
        self.logger.info('Generating DROP output...')

        # Set up dask cluster
        self._set_dask()

        if isinstance(self.var, list):
            for var in self.var:
                self._write_var(var)

        else:  # Only one variable
            self._write_var(self.var)

        self.logger.info('Move tmp files from %s to output directory %s', self.tmpdir, self.outdir)
        # Move temporary files to output directory
        move_tmp_files(self.tmpdir, self.outdir)

        # Cleaning
        self.data.close()
        self._close_dask()
        self._remove_tmpdir()

        self.logger.info('Finished generating DROP output.')

    def _define_source_grid_name(self):
        """"
        Define the source grid name based on the resolution
        """
        if self.resolution in self.default_grids:
            return 'lon-lat'
        if self.resolution == 'native':
            try:
                return self.reader.source_grid_name
            except AttributeError:
                self.logger.warning('No source grid name defined in the reader, using resolution as source grid name')
                return False
        return self.resolution

    def create_catalog_entry(self):
        """
        Create an entry in the catalog for DROP
        """
        # find the catalog of my experiment and load it
        catalogfile = os.path.join(self.configdir, 'catalogs', self.catalog,
                                   'catalog', self.model, self.exp + '.yaml')
        
        with SafeFileLock(catalogfile + '.lock', loglevel=self.loglevel):
            cat_file = load_yaml(catalogfile)

            # define the entry name
            entry_name = self.catbuilder.create_entry_name()
            sgn = self._define_source_grid_name()

            if entry_name in cat_file['sources']:
                catblock = cat_file['sources'][entry_name]
            else:
                catblock = None

            block = self.catbuilder.create_entry_details(
                basedir=self.basedir, catblock=catblock, 
                source_grid_name=sgn
            )

            cat_file['sources'][entry_name] = block

            # dump the update file
            dump_yaml(outfile=catalogfile, cfg=cat_file)

    def create_zarr_entry(self, verify=True):
        """
        Create a Zarr entry in the catalog for DROP

        Args:
            verify: open the DROP source and verify it can be read by the reader
        """
        full_dict, partial_dict = list_drop_files_complete(self.outdir)

        # extra zarr only directory
        zarrdir = os.path.join(self.outdir, 'zarr')
        create_folder(zarrdir)

        # this dictionary based structure is an overkill but guarantee flexibility
        urlpath = []
        for key, value in full_dict.items():
            jsonfile = os.path.join(zarrdir, f'drop-yearly-{key}.json')
            self.logger.debug('Creating zarr files for full files %s', key)
            if value:
                jsonfile = create_zarr_reference(value, jsonfile, loglevel=self.loglevel)
                if jsonfile is not None:
                    urlpath = urlpath + [f'reference::{jsonfile}']

        for key, value in partial_dict.items():
            jsonfile = os.path.join(zarrdir, f'drop-monthly-{key}.json')
            self.logger.debug('Creating zarr files for partial files %s', key)
            if value:
                jsonfile = create_zarr_reference(value, jsonfile, loglevel=self.loglevel)
                if jsonfile is not None:
                    urlpath = urlpath + [f'reference::{jsonfile}']

        if not urlpath:
            raise FileNotFoundError('No files found to create zarr reference')

        # apply intake replacement: works on string need to loop on the list
        for index, value in enumerate(urlpath):
            urlpath[index] = replace_intake_vars(catalog=self.catalog, path=value)

        # find the catalog of my experiment and load it
        catalogfile = os.path.join(self.configdir, 'catalogs', self.catalog,
                                   'catalog', self.model, self.exp + '.yaml')
        
        with SafeFileLock(catalogfile + '.lock', loglevel=self.loglevel):
            cat_file = load_yaml(catalogfile)

            # define the entry name - zarr entries never have lra- prefix
            base_name = f'{self.catbuilder.resolution}-{self.catbuilder.frequency}'
            entry_name = base_name + '-zarr'
            self.logger.info('Creating zarr files for %s %s %s', self.model, self.exp, entry_name)
            sgn = self._define_source_grid_name()

            if entry_name in cat_file['sources']:
                catblock = cat_file['sources'][entry_name]
            else:
                catblock = None

            block = self.catbuilder.create_entry_details(
                basedir=self.basedir, catblock=catblock, source_grid_name=sgn, driver='zarr'
            )
            block['args']['urlpath'] = urlpath
            cat_file['sources'][entry_name] = block

            dump_yaml(outfile=catalogfile, cfg=cat_file)

        # verify the zarr entry makes sense
        if verify:
            self.logger.info('Verifying that zarr entry can be loaded...')
            try:
                reader = Reader(model=self.model, exp=self.exp, source=entry_name)
                _ = reader.retrieve()
                self.logger.info('Zarr entry successfully created!!!')
            except (KeyError, ValueError) as e:
                self.logger.error('Cannot load zarr DROP with error --> %s', e)
                self.logger.error('Zarr source is not accessible by the Reader likely due to irregular amount of NetCDF file')
                self.logger.error('To avoid issues in the catalog, the entry will be removed')
                self.logger.error('In case you want to keep it, please run with verify=False')
                with SafeFileLock(catalogfile + '.lock', loglevel=self.loglevel):
                    cat_file = load_yaml(catalogfile)
                    del cat_file['sources'][entry_name]
                    dump_yaml(outfile=catalogfile, cfg=cat_file)

    def _set_dask(self):
        """
        Set up dask cluster
        """
        if self.dask:  # self.nproc > 1
            self.logger.info('Setting up dask cluster with %s workers', self.nproc)
            dask.config.set({'temporary_directory': self.tmpdir})
            self.logger.info('Temporary directory: %s', self.tmpdir)
            self.cluster = LocalCluster(n_workers=self.nproc,
                                        threads_per_worker=1)
            self.client = Client(self.cluster)
        else:
            self.client = None
            dask.config.set(scheduler='synchronous')

    def _close_dask(self):
        """
        Close dask cluster
        """
        if self.dask:  # self.nproc > 1
            self.client.shutdown()
            self.cluster.close()
            self.logger.info('Dask cluster closed')

    def _remove_tmpdir(self):
        """
        Remove temporary directory
        """
        self.logger.info('Removing temporary directory %s', self.tmpdir)
        shutil.rmtree(self.tmpdir)

    def _concat_var_year(self, var, year):
        """
        To reduce the amount of files concatenate together all the files
        from the same year
        """

        infiles_pattern = self.get_filename(var, year, month='??')
        monthly_files = sorted(glob.glob(infiles_pattern))

        if len(monthly_files) == 12:
            self.logger.info('Creating a single file for %s, year %s...', var, str(year))
            outfile = self.get_filename(var, year)
            tmp_outfile = self.get_filename(var, year, tmp=True)
            
            # Move monthly files to tmp for safety
            for monthly_file in monthly_files:
                shutil.move(monthly_file, self.tmpdir)
            
            # Clean any existing output files
            for f in [tmp_outfile, outfile]:
                if os.path.exists(f):
                    os.remove(f)

            # Get the moved files in tmpdir - they keep the same basename
            tmp_monthly_files = [os.path.join(self.tmpdir, os.path.basename(f)) for f in monthly_files]

            # Concatenation with CDO or Xarray
            if self.compact == 'cdo':
                command = [
                    'cdo',
                    *self.cdo_options,
                    'cat',
                    *tmp_monthly_files,
                    tmp_outfile
                ]
                self.logger.debug("Using CDO command: %s", command)
                subprocess.check_output(command, stderr=subprocess.STDOUT)
            else:
                self.logger.debug("Using xarray to concatenate files")
                xfield = xr.open_mfdataset(tmp_monthly_files, combine='by_coords', parallel=True)
                name = list(xfield.data_vars)[0]
                xfield.to_netcdf(tmp_outfile,
                                 encoding={'time': self.time_encoding, name: self.var_encoding})

            # Move back the yearly file and cleanup
            shutil.move(tmp_outfile, outfile)
            for tmp_file in tmp_monthly_files:
                self.logger.info('Cleaning %s...', tmp_file)
                os.remove(tmp_file)

    def get_filename(self, var, year=None, month=None, tmp=False):
        """Create output filenames"""

        filename = self.outbuilder.build_filename(var=var, year=year, month=month)

        if tmp:
            filename = os.path.join(self.tmpdir, filename)
        else:
            filename = os.path.join(self.outdir, filename)

        return filename

    def check_integrity(self, varname):
        """To check if the DROP entry is fine before running"""

        yearfiles = self.get_filename(varname)
        yearfiles = glob.glob(yearfiles)
        checks = [file_is_complete(yearfile, loglevel=self.loglevel) for yearfile in yearfiles]
        all_checks_true = all(checks) and len(checks) > 0
        if all_checks_true and not self.overwrite:
            self.logger.info('All the data produced seems complete for var %s...', varname)
            last_record = xr.open_mfdataset(self.get_filename(varname)).time[-1].values
            self.last_record = pd.to_datetime(last_record).strftime('%Y%m%d')
            self.check = True
            self.logger.info('Last record archived is %s...', self.last_record)
        else:
            self.check = False
            self.logger.warning('Still need to run for var %s...', varname)

    def _write_var(self, var):
        """Call write var for generator or catalog access"""
        t_beg = time()

        self._write_var_catalog(var)

        t_end = time()
        self.logger.info('Process took %.4f seconds', t_end - t_beg)

    def _remove_regridded(self, data):

        # remove regridded attribute to avoid issues with Reader
        # https://github.com/oloapinivad/AQUA/issues/147
        if "AQUA_regridded" in data.attrs:
            self.logger.debug('Removing regridding attribute...')
            del data.attrs["AQUA_regridded"]
        return data

    def _write_var_catalog(self, var):
        """
        Write variable to file

        Args:
            var (str): variable name
        """

        self.logger.info('Processing variable %s...', var)
        temp_data = self.data[var]

        if self.frequency:
            temp_data = self.reader.timstat(temp_data, self.stat, freq=self.frequency,
                                            exclude_incomplete=self.exclude_incomplete)

        # temp_data could be empty after time statistics if everything was excluded
        if 'time' in temp_data.coords and len(temp_data.time) == 0:
            self.logger.warning('No data available for variable %s after time statistics, skipping...', var)
            return
        
        # regrid
        if self.resolution:
            temp_data = self.reader.regrid(temp_data)
            temp_data = self._remove_regridded(temp_data)

        if self.region:
            temp_data = self.reader.select_area(temp_data, lon=self.region['lon'], lat=self.region['lat'], drop=self.drop)

        # Splitting data into yearly files
        years = sorted(set(temp_data.time.dt.year.values))
        if self.performance_reporting:
            years = [years[0]]
        for year in years:

            self.logger.info('Processing year %s...', str(year))
            yearfile = self.get_filename(var, year=year)

            # checking if file is there and is complete
            filecheck = file_is_complete(yearfile, loglevel=self.loglevel)
            if filecheck:
                if not self.overwrite:
                    self.logger.info('Yearly file %s already exists, skipping...', yearfile)
                    continue
                self.logger.warning('Yearly file %s already exists, overwriting as requested...', yearfile)
            year_data = temp_data.sel(time=temp_data.time.dt.year == year)

            # Splitting data into monthly files
            months = sorted(set(year_data.time.dt.month.values))
            if self.performance_reporting:
                months = [months[0]]
            for month in months:
                self.logger.info('Processing month %s...', str(month))
                outfile = self.get_filename(var, year=year, month=month)

                # checking if file is there and is complete
                filecheck = file_is_complete(outfile, loglevel=self.loglevel)
                if filecheck:
                    if not self.overwrite:
                        self.logger.info('Monthly file %s already exists, skipping...', outfile)
                        continue
                    self.logger.warning('Monthly file %s already exists, overwriting as requested...', outfile)

                month_data = year_data.sel(time=year_data.time.dt.month == month)

                # real writing
                if self.definitive:
                    tmpfile = self.get_filename(var, year=year, month=month, tmp=True)
                    schunk = time()
                    self.write_chunk(month_data, tmpfile)
                    tchunk = time() - schunk
                    self.logger.info('Chunk execution time: %.2f', tchunk)

                    # check everything is correct
                    filecheck = file_is_complete(tmpfile, loglevel=self.loglevel)
                    # we can later add a retry
                    if not filecheck:
                        self.logger.error('Something has gone wrong in %s!', tmpfile)
                    self.logger.info('Moving temporary file %s to %s', tmpfile, outfile)

                    move_tmp_files(self.tmpdir, self.outdir)
                del month_data
            del year_data
            if self.definitive and self.compact:
                self._concat_var_year(var, year)
        del temp_data

    def append_history(self, data):
        """
        Append comprehensive processing history to the data attributes
               
        Args:
            data: xarray Dataset or DataArray to append history to
            
        Returns:
            data: Input data with updated history attribute
        """
        history_list = ["DROP"]
        
        # Add regridding information
        if self.resolution:
            history_list.append(f"regridded from {self.reader.src_grid_name} to {self.resolution}")
        if self.frequency and self.stat:
            history_list.append(
                f"resampled from frequency {self.reader.timemodule.orig_freq} to {self.frequency} "
                f"using {self.stat} statistic")
        if self.region and self.region_name:
            region_info = f"regional selection applied ({self.region_name})"
            history_list.append(region_info)
        
        # Build the complete sentence
        if len(history_list) == 1:
            history = history_list[0]
        else:
            history = history_list[0] + ": " + ", ".join(history_list[1:])

        log_history(data, history)

        return data

    def write_chunk(self, data, outfile):
        """Write a single chunk of data - Xarray Dataset - to a specific file
        using dask if required and monitoring the progress"""

        data = self.append_history(data)

        # File to be written
        if os.path.exists(outfile):
            os.remove(outfile)
            self.logger.warning('Overwriting file %s...', outfile)

        self.logger.info("Computing to write file %s...", outfile)

        # Compute + progress monitoring
        if self.dask:
            if self.performance_reporting:
                # Full Dask dashboard to HTML
                filename = f"dask-{self.model}-{self.exp}-{self.source}-{self.nproc}.html"
                with performance_report(filename=filename):
                    job = data.persist()
                    progress(job)
                    job = job.compute()
            else:
                # Memory monitoring always on
                ms = MemorySampler()
                with ms.sample("chunk"):
                    job = data.persist()
                    progress(job)
                    job = job.compute()
                array_data = np.array(ms.samples["chunk"])
                avg_mem = np.mean(array_data[:, 1]) / 1e9
                max_mem = np.max(array_data[:, 1]) / 1e9
                self.logger.info("Avg memory used: %.2f GiB, Peak memory used: %.2f GiB", avg_mem, max_mem)
        else:
            with ProgressBar():
                job = data.compute()

        # Final safe NetCDF write (serial, no dask)
        job.to_netcdf(
            outfile,
            encoding={"time": self.time_encoding, data.name: self.var_encoding},
        )
        del job
        self.logger.info('Writing file %s successful!', outfile)



import xarray as xr
import matplotlib.pyplot as plt

from aqua.core.logger import log_configure
from aqua.core.util import to_list, coord_names
from .styles import ConfigStyle

def plot_lat_lon_profiles(data: xr.DataArray | list[xr.DataArray],
                          ref_data: xr.DataArray | None = None,
                          ref_std_data: xr.DataArray | None = None,
                          data_labels: list | None = None,
                          ref_label: str | None = None,
                          style: str | None = None,
                          fig: plt.Figure | None = None,
                          ax: plt.Axes | None = None,
                          figsize: tuple = (10, 5),
                          title: str | None = None,
                          loglevel: str = 'WARNING'):
    """
    Plot latitude or longitude profiles of data, averaging over the specified axis.

    Args:
        data (xr.DataArray | list[xr.DataArray] | None): Data to plot. Must be xarray DataArrays 
            with 'lat', 'lon', 'latitude', or 'longitude' dimensions. Can be a single DataArray 
            or a list of DataArrays.
        ref_data (xr.DataArray, optional): Reference data to plot.
        ref_std_data (xr.DataArray | None, optional): Standard deviation of the reference data.
        data_labels (list | None, optional): Labels for the data.
        ref_label (str | None, optional): Label for the reference data.
        style (str | None, optional): Style for the plot.
        fig (plt.Figure | None, optional): Matplotlib figure object.
        ax (plt.Axes | None, optional): Matplotlib axes object.
        figsize (tuple, optional): Figure size if a new figure is created.
        title (str | None, optional): Title for the plot.
        loglevel (str, optional): Logging level.

    Returns:
        tuple: Matplotlib figure and axes objects.
    """
    logger = log_configure(loglevel, 'plot_lat_lon_profiles')
    ConfigStyle(style=style, loglevel=loglevel)
    
    # Convert data to list, handling both single DataArrays and lists
    data_list = to_list(data)

    # Create figure if needed
    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot(1, 1, 1)

    logger.debug(f"Plotting {len(data_list)} data arrays")

    # Plot data (higher zorder = foreground)
    for i, d in enumerate(data_list):
        # Determine coordinate name based on 'lat' or 'lon'
        lon_name, lat_name = coord_names(d)
        coord_name = lat_name if lat_name is not None else lon_name

        if coord_name is None:
            logger.warning(f"Data {i} has no spatial coordinates, skipping")
            continue
            
        label = data_labels[i] if data_labels else None
        d.plot(ax=ax, label=label, linewidth=3, zorder=3)

    # Handle reference data (lower zorder = background)
    if ref_data is not None:
        
        # Find coordinate for ref_data
        lon_name, lat_name = coord_names(ref_data)
        coord_name = lat_name if lat_name is not None else lon_name
        
        if coord_name is not None:
            ref_x_coord = ref_data[coord_name].values
            
            # Plot reference std if available
            if ref_std_data is not None:
                if hasattr(ref_std_data, 'compute'):
                    ref_std_data = ref_std_data.compute()
                ax.fill_between(ref_x_coord,
                                ref_data.values - 2.*ref_std_data.values,
                                ref_data.values + 2.*ref_std_data.values,
                                facecolor='grey', alpha=0.5, zorder=1)

            # Plot reference data
            ref_data.plot(ax=ax, label=ref_label if ref_label else 'Reference',
                          color='black', linestyle='-', linewidth=3, alpha=1.0,
                          zorder=2)

    # Finalize plot
    if data_labels:
        ax.legend(fontsize='small')
    ax.grid(True, axis="y", linestyle='-', color='silver', alpha=0.8)

    # Set x-label based on the first valid coordinate found
    first_data = data_list[0]
    lon_name, lat_name = coord_names(first_data)
    coord_name = lat_name if lat_name is not None else lon_name

    if coord_name and 'lat' in coord_name:
        ax.set_xlabel('Latitude')
    elif coord_name and 'lon' in coord_name:
        ax.set_xlabel('Longitude')

    # Set title if provided
    if title:
        ax.set_title(title, fontsize=13, fontweight='bold')

    return fig, ax


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
AQUA analysis checker command line interface.
Check that the imports are correct and the requested model is available in the
Reader catalog.
'''
import sys
import os
import argparse
from aqua.core.util import template_parse_arguments


def parse_arguments(args):
    """
    Parse command line arguments

    Args:
        args: list of arguments

    Returns:
        argparse.Namespace: parsed arguments
    """
    parser = argparse.ArgumentParser(description='Check setup CLI')
    parser = template_parse_arguments(parser)

    parser.add_argument("--yaml", type=str,
                        required=False, help="write an experiment.yaml file to a given directory")
    parser.add_argument("--no-read", action="store_false", dest='read',
                        required=False, help="do not attempt to read data (used with --yaml to speed up when creating only yaml)")
    parser.add_argument("--no-rebuild", action="store_false", dest='rebuild',
                        required=False, help="by default rebuild of areas is forced, this prevents it")

    return parser.parse_args(args)


if __name__ == '__main__':

    args = parse_arguments(sys.argv[1:])

    try:
        from aqua import __version__ as aqua_version
        from aqua import Reader
        from aqua.core.logger import log_configure
        from aqua.core.util import get_arg, dump_yaml
        from aqua.core.exceptions import NoDataError

        loglevel = get_arg(args, 'loglevel', 'WARNING')
        logger = log_configure(log_name='Setup check', log_level=loglevel)
    except ImportError:
        raise ImportError('Failed to import aqua. Check that you have installed aqua.'
                          'If you are using a conda environment, check that you have activated it.'
                          'If you are using a container, check that you have started it.')
    except Exception as e:
        raise ImportError('Failed to import aqua: {}'.format(e))

    logger.info("Running Setup Checker with AQUA version %s", aqua_version)

    catalog = get_arg(args, 'catalog', None)
    model = get_arg(args, 'model', None)
    exp = get_arg(args, 'exp', None)
    source = get_arg(args, 'source', None)
    # For some diagnostics the regrid is required, so we set a default
    # which is 1 degree for the regrid. The user can override it
    # with the --regrid argument.
    regrid = get_arg(args, 'regrid', 'r100')
    realization = get_arg(args, 'realization', None)
    reader_kwargs = {'realization': realization} if realization else {}
    yamldir = get_arg(args, 'yaml', None)
    fread = getattr(args, 'read', True)       # --no-read means fread=False, default is to read data
    frebuild = getattr(args, 'rebuild', True) # --no-rebuild means frebuild=False, default is to rebuild areas

    if model is None or exp is None or source is None:
        raise ValueError('model, exp and source are required arguments')
    if catalog is None:
        logger.warning('No catalog provided, determining the catalog with the Reader')

    try:
        reader = Reader(catalog=catalog, model=model, exp=exp, source=source,
                        loglevel=loglevel, rebuild=frebuild, regrid=regrid,
                        **reader_kwargs)

        # extract metadata from catalog
        if yamldir:
            logger.info('Creating experiment.yaml')
            metadata = reader.expcat.metadata.copy()
            metadata.pop("catalog_dir", None)
            metadata['description'] = getattr(reader.expcat, 'description', '')
            metadata['catalog'] = catalog
            metadata['model'] = model
            metadata['experiment'] = exp
            yaml_path = os.path.join(yamldir, "experiment.yaml")
            dump_yaml(outfile=yaml_path, cfg=metadata)

        if fread:
            reader.retrieve(sample=True)

    except Exception as e:
        logger.error('Failed to retrieve data: {}'.format(e))
        logger.error('Check that the model is available in the Reader catalog.')
        raise NoDataError('Failed to retrieve data: {}'.format(e))

    logger.info('Check is terminated, diagnostics can run!')



"""Module for scientific utility functions."""
import xarray as xr

# set default options for xarray
xr.set_options(keep_attrs=True)


TRIPLET_MONTHS = {
    'DJF': [12, 1, 2],   # December-January-February
    'JFM': [1, 2, 3],    # January-February-March
    'FMA': [2, 3, 4],    # February-March-April
    'MAM': [3, 4, 5],    # March-April-May
    'AMJ': [4, 5, 6],    # April-May-June
    'MJJ': [5, 6, 7],    # May-June-July
    'JJA': [6, 7, 8],    # June-July-August
    'JAS': [7, 8, 9],    # July-August-September
    'ASO': [8, 9, 10],   # August-September-October
    'SON': [9, 10, 11],  # September-October-November
    'OND': [10, 11, 12], # October-November-December
    'NDJ': [11, 12, 1],  # November-December-January
}

def lon_to_360(lon: float) -> float:
    """
    Convert longitude from [-180,180] (or any value) to [0,360].

    Args:
        lon (float): longitude coordinate

    Returns:
        float: converted longitude
    """
    return 360 if lon == 360 else lon % 360


def lon_to_180(lon: float) -> float:
    """
    Convert longitude from [0,360] (or any value) to [-180,180].

    Args:
        lon (float): longitude coordinate

    Returns:
        float: converted longitude
    """
    if lon == -180:
        return -180
    lon = lon % 360
    return lon - 360 if lon > 180 else lon


def check_coordinates(lon: list | None, lat: list | None,
                           default_coords: dict) -> tuple[list, list]:
        """
        Validate and normalize latitude/longitude ranges.

        Returns:
            tuple: (lon_range, lat_range) with values mapped to default system.
        """
        # --- Latitude ---
        # Populate with maximum extent if no Latitude is provided
        if lat is None:
            lat = [default_coords["lat_min"], default_coords["lat_max"]]
        # Swap if values are inverted
        elif lat[0] > lat[1]:
            lat = [lat[1], lat[0]]
        # If the two latitudes are equal raise an error
        elif lat[0] == lat[1]:
            raise ValueError(f"Both latitude values are equal: {lat[0]}, please provide a valid range.")
        # Check that values are within the maximum range
        if lat[0] < default_coords["lat_min"] or lat[1] > default_coords["lat_max"]:
            raise ValueError(f"Latitude must be within {default_coords['lat_min']} and {default_coords['lat_max']}")

        # --- Longitude ---
        # Populate with maximum extent if no Longitude is provided
        if lon is None or (lon[0] == 0 and lon[1] == 360) or (lon[0] == -180 and lon[1] == 180):
            lon = [default_coords["lon_min"], default_coords["lon_max"]]
        # If the two longitudes are equal raise an error
        elif lon[0] == lon[1]:
            raise ValueError(f"Longitude: {lon[0]} == {lon[1]}, please provide a valid range.")
        else:
            # Normalize according to coordinate system
            if default_coords["lon_min"] == 0 and default_coords["lon_max"] == 360:
                lon = [lon_to_360(l) for l in lon]
            elif default_coords["lon_min"] == -180 and default_coords["lon_max"] == 180:
                lon = [lon_to_180(l) for l in lon]

        return lon, lat


def select_season(xr_data, season: str):
    """
    Select a season from a xarray.DataArray or xarray.Dataset.

    Args:
        xr_data (xarray.DataArray or xarray.Dataset): input data
        season (str): season to be selected. Available options are defined in TRIPLET_MONTHS.
    Returns:
        (xarray.DataArray or xarray.Dataset): selected season
    """
    if season in TRIPLET_MONTHS:
        selected_months = TRIPLET_MONTHS[season]
        selected =  xr_data.sel(time=(xr_data['time.month'] == selected_months[0]) | (xr_data['time.month'] == selected_months[1]) | (xr_data['time.month'] == selected_months[2]))
        # Add AQUA_season attribute
        selected.attrs['AQUA_season'] = season
        return selected
    elif season == 'annual':
        return xr_data
    else:
        raise ValueError(f"Invalid season abbreviation. Available options are: {', '.join(TRIPLET_MONTHS.keys())}, or 'annual' to perform no season selection.")


def generate_quarter_months(anchor_month='JAN'):
    """
    Construct four consecutive quarters every 3rd triplet starting from the anchor month.
    Args:
        anchor_month (str): The anchor month for the quarterly groupings.
    Returns:
        dict: A dictionary of quarterly month groupings.
    """
    anchor_month = anchor_month.upper()

    monlist = ['JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN', 'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC']
    triplet_keys = list(TRIPLET_MONTHS)

    if anchor_month not in monlist:
        raise ValueError(f"Invalid anchor month: {anchor_month}. Must be one of: {', '.join(monlist)}")

    anchor_month_num = monlist.index(anchor_month) + 1

    start_idx_anchor_month = next(i for i, key in enumerate(triplet_keys) 
                                  if TRIPLET_MONTHS[key][0] == anchor_month_num)

    quarter_months = {f"Q{q + 1}": TRIPLET_MONTHS[triplet_keys[(start_idx_anchor_month + q * 3) % len(triplet_keys)]] for q in range(4)}

    return {anchor_month: quarter_months}
    

def merge_attrs(target, source, overwrite=False):
    """Merge attributes from source into target.

    Args:
        target (xr.Dataset or xr.DataArray or dict): The target for merging.
        source (xr.Dataset or xr.DataArray or dict): The source of attributes.
        overwrite (bool): If True, overwrite existing keys in target.
                          If False, only add keys that don't already exist.
    """
    if isinstance(target, (xr.Dataset, xr.DataArray)):
        target = target.attrs
    if isinstance(source, (xr.Dataset, xr.DataArray)):
        source = source.attrs

    for k, v in source.items():
        if overwrite or k not in target:
            target[k] = v


def find_vert_coord(ds):
    """
    Identify the vertical coordinate name(s) based on coordinate units. Returns always a list.
    The list will be empty if none found.
    """
    vert_coord = [x for x in ds.coords if ds.coords[x].attrs.get("units") in ["Pa", "hPa", "m", "km", "Km", "cm", ""]]
    return vert_coord



import pytest
import regionmask
import xarray as xr
import numpy as np
from typeguard import TypeCheckError
from aqua.core.fldstat import AreaSelection
from aqua.core.util import select_season
from aqua.core.util.sci_util import generate_quarter_months
from aqua.core.util import check_seasonal_chunk_completeness

from conftest import LOGLEVEL

loglevel = LOGLEVEL

@pytest.fixture
def sample_data():
    """Create a sample DataArray for testing"""
    data = xr.DataArray(
        np.random.rand(6, 6),
        coords={'lat': [10, 20, 30, 40, 50, 60], 'lon': [40, 50, 60, 70, 80, 90]},
        dims=['lat', 'lon']
    )
    return data

@pytest.mark.aqua
def test_valid_selection_no_brd(sample_data):
    """Test with valid latitude and longitude ranges, no box_brd"""
    lat_range = [10, 30]
    lon_range = [45, 55]
    result = AreaSelection(loglevel=loglevel).select_area(sample_data,
                                                          lat=lat_range, lon=lon_range,
                                                          box_brd=False)

    assert result is not None
    assert np.isnan(result.sel(lat=10, lon=40).values)
    assert np.isnan(result.sel(lat=60, lon=90).values)


@pytest.mark.aqua
def test_selection_regionmask(sample_data):
    """Test selection using regionmask"""
    # Define a simple regionmask region
    region = regionmask.defined_regions.natural_earth_v5_0_0.countries_110
    region_sel = ['United States of America', 'Russia']

    result = AreaSelection(loglevel=loglevel).select_area(sample_data,
                                                          region=region,
                                                          region_sel=region_sel,
                                                          box_brd=True)

    assert result is not None
    # Check that values outside the selected regions are NaN
    assert np.isnan(result.sel(lat=0, lon=150, method="nearest").values)
    # Check that values inside the selected regions are not NaN
    assert result.sel(lat=40, lon=-100, method="nearest").values is not np.nan


@pytest.mark.aqua
def test_valid_selection(sample_data):
    """Test with valid latitude and longitude ranges"""
    lat_range = [15, 25]
    lon_range = [45, 55]
    result = AreaSelection(loglevel=loglevel).select_area(sample_data,
                                                          lat=lat_range, lon=lon_range,
                                                          box_brd=True)

    assert result is not None
    assert np.isnan(result.sel(lat=10, lon=40).values)
    assert np.isnan(result.sel(lat=60, lon=90).values)


@pytest.mark.aqua
def test_missing_lat_lon(sample_data):
    """Test when both lat and lon are None, the sample_data is returned"""
    result = AreaSelection(loglevel=loglevel).select_area(sample_data, lat=None, lon=None, box_brd=True)
    assert result is not None
    assert result.equals(sample_data)


@pytest.mark.aqua
def test_missing_lat_lon_coords():
    """Test with missing lat or lon coordinates, should raise an KeyError"""
    data_missing_lat = xr.DataArray(np.random.rand(3, 3),
                                    coords={'lon': [40, 50, 60]},
                                    dims=['lat', 'lon'])
    data_missing_lon = xr.DataArray(np.random.rand(3, 3),
                                    coords={'lat': [10, 20, 30]},
                                    dims=['lat', 'lon'])

    with pytest.raises(KeyError):
        AreaSelection(loglevel=loglevel).select_area(data_missing_lat, lat=[15, 25], lon=[45, 55],
                                                      box_brd=True)

    with pytest.raises(KeyError):
        AreaSelection(loglevel=loglevel).select_area(data_missing_lon, lat=[15, 25], lon=[45, 55],
                                                      box_brd=True)


@pytest.mark.aqua
def test_missing_data():
    """Test with missing data or wrong type"""
    with pytest.raises(TypeError):
        AreaSelection(loglevel=loglevel).select_area(lat=[15, 25], lon=[45, 55],
                                                      box_brd=True)
        
    with pytest.raises(TypeCheckError):
        AreaSelection(loglevel=loglevel).select_area('invalid_data', lat=[15, 25], lon=[45, 55],
                                                      box_brd=True)


@pytest.mark.aqua
def test_select_season():
    """Test the select_season function with valid and invalid inputs."""
    # Sample DataArray with a time dimension
    times = xr.date_range(start="2000-01-01", periods=12, freq="MS")
    data = xr.DataArray(np.random.rand(12), coords={"time": times}, dims=["time"])

    # Valid season tests
    for season, expected_months in {"DJF": [12, 1, 2], "MAM": [3, 4, 5]}.items():
        result = select_season(data, season)
        assert len(result) == 3
        assert all(month in expected_months for month in result["time.month"].values)

    # Invalid season test
    with pytest.raises(ValueError):
        select_season(data, "XYZ")

    # Missing time dimension test
    data_no_time = xr.DataArray(np.random.rand(12), dims=["dim_0"])
    with pytest.raises(KeyError):
        select_season(data_no_time, "DJF")


@pytest.mark.aqua
def test_generate_quarter_months():
    """Test the generate_quarter_months function with various anchor months."""
    result = generate_quarter_months('DEC')
    assert result == {'DEC': {'Q1': [12, 1, 2], 'Q2': [3, 4, 5], 'Q3': [6, 7, 8], 'Q4': [9, 10, 11]}}

    result = generate_quarter_months('MAR')
    assert result == {'MAR': {'Q1': [3, 4, 5], 'Q2': [6, 7, 8], 'Q3': [9, 10, 11], 'Q4': [12, 1, 2]}}
    with pytest.raises(ValueError):
        generate_quarter_months('XXX')


@pytest.mark.aqua
def test_check_seasonal_chunk_completeness():
    # Daily data:
    t_dec = xr.date_range("2000-12-01", "2000-12-31", freq="D") # 2000-12 (present)
    t_feb = xr.date_range("2001-02-01", "2001-02-28", freq="D") # 2001-01 (missing); DJF incomplete
    t_mam = xr.date_range("2001-03-01", "2001-05-31", freq="D") # 2001-03..2001-05; MAM complete
    time = t_dec.append(t_feb).append(t_mam)

    da = xr.DataArray(np.ones(time.size), coords={"time": time}, dims=["time"])
    mask = check_seasonal_chunk_completeness(da, resample_frequency="QS-DEC", loglevel="DEBUG")

    # Expected seasonal chunk start times under QS-DEC in this range: 2000-12-01 and 2001-03-01
    assert "2000-12-01" in mask.time.dt.strftime("%Y-%m-%d").values
    assert "2001-03-01" in mask.time.dt.strftime("%Y-%m-%d").values

    # DJF (starts 2000-12-01) is incomplete (missing January)
    assert bool(mask.sel(time="2000-12-01").item()) is False
    # MAM (starts 2001-03-01) is complete (Mar, Apr, May present)
    assert bool(mask.sel(time="2001-03-01").item()) is True



"""DROP module."""
from .drop import Drop
from .output_path_builder import OutputPathBuilder
from .catalog_entry_builder import CatalogEntryBuilder

__all__ = ['Drop', 'OutputPathBuilder', 'CatalogEntryBuilder']



from .coordidentifier import CoordIdentifier
from .coordtransformer import CoordTransformer, counter_reverse_coordinate
from .datamodel import DataModel

__all__ = [
    "CoordIdentifier",
    "CoordTransformer",
    "DataModel",
    "counter_reverse_coordinate"
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
AQUA installation operations mixin - includes also listing
'''

import os
import shutil
import sys

from aqua.core.lock import SafeFileLock
from aqua.core.util import load_yaml, dump_yaml
from aqua.core.configurer import ConfigPath

from .util import query_yes_no

# check if aqua.diagnostics is installed
try:
    from aqua.diagnostics import DIAGNOSTIC_CONFIG_DIRECTORIES
    from aqua.diagnostics import DIAGNOSTIC_TEMPLATE_DIRECTORIES
except ImportError:
    DIAGNOSTIC_CONFIG_DIRECTORIES = []
    DIAGNOSTIC_TEMPLATE_DIRECTORIES = []

# folder used for reading/storing catalogs
CATPATH = 'catalogs'

# directories to be installed in the AQUA config folder
CORE_CONFIG_DIRECTORIES = ['catgen', 'data_model', 'fixes', 'grids', 'styles']
CORE_TEMPLATE_DIRECTORIES = ['catgen', 'drop', 'gridbuilder']


class InstallMixin:
    """Mixin for AQUA installation operations"""

    def _check(self, silent=False, return_info=False):
        """
        Check installation and optionally return detailed information
        
        Args:
            silent (bool): If True, suppress errors and use ERROR loglevel
            return_info (bool): If True, return installation info dict instead of exiting
            
        Returns:
            dict or None: If return_info=True, returns:
                {
                    'installed': bool,
                    'configpath': str or None,
                    'core': {'installed': bool, 'mode': str},
                    'diagnostics': {'installed': bool, 'mode': str}
                }
        """
        checklevel = 'ERROR' if silent else self.loglevel

        try:
            self.configpath = ConfigPath(loglevel=checklevel).configdir
            self.configfile = os.path.join(self.configpath, 'config-aqua.yaml')
            self.templatepath = os.path.join(self.configpath, 'templates')
            self.logger.debug('AQUA found in %s', self.configpath)

            if return_info:
                # Gather detailed installation information
                info = {
                    'installed': True,
                    'configpath': self.configpath,
                    'core': {
                        'installed': self._check_component_installed('core'),
                        'mode': self._get_component_mode('core')
                    },
                    'diagnostics': {
                        'installed': self._check_component_installed('diagnostics'),
                        'mode': self._get_component_mode('diagnostics')
                    }
                }
                return info
                
        except FileNotFoundError:
            if return_info:
                return {
                    'installed': False,
                    'configpath': None,
                    'core': {'installed': False, 'mode': 'not_installed'},
                    'diagnostics': {'installed': False, 'mode': 'not_installed'}
                }
            
            self.logger.error('No AQUA installation found!')
            sys.exit(1)
        
    def _check_component_installed(self, component):
        """
        Check if a specific component is already installed
        
        Args:
            component (str): 'core' or 'diagnostics'
            
        Returns:
            bool: True if component is installed
        """

        # get the first config directory for the component
        directories = self._get_config_dirs(component)
        for directory in directories:
            if os.path.exists(os.path.join(self.configpath, directory)):
                return True
        return False

    def _get_component_mode(self, component):
        """
        Check if a component is installed in editable mode
        
        Args:
            component (str): 'core' or 'diagnostics'
            
        Returns:
            str: 'editable', 'standard', or 'not_installed'
        """
        # Check the installation mode based on the first config directory
        directories = self._get_config_dirs(component)
        for directory in directories:
            path = os.path.join(self.configpath, directory)
            if os.path.islink(path):
                return 'editable'
            if os.path.exists(path):
                return 'standard'

        return None

    def _validate_component_path(self, component, custom_path):
        """Validate that editable install path exists and return the source path
        
        Args:
            component (str): 'core' or 'diagnostics'
            custom_path (str): Path to the editable installation
            
        Returns:
            str: Absolute path to the component config directory
        """
        actual_source = os.path.join(
            os.path.abspath(custom_path), 'aqua', component, 'config'
        )
        if not os.path.exists(actual_source):
            self.logger.error(f'{actual_source} does not exist for {component}')
            sys.exit(1)
        return actual_source

    def _install_component(self, component, mode, source_path, config_dirs, template_dirs):
        """
        Install a component (core or diagnostics) in either standard or editable mode.
        
        Args:
            component (str): 'core' or 'diagnostics'
            mode (dict): Installation mode with 'mode' and 'path' keys
            source_path (str): Default source path for the component
            config_dirs (list): List of config directories to install
            template_dirs (list): List of template directories to install
        """
        install_mode = mode['mode']
        custom_path = mode['path']

        self.logger.info(f'Installing {component} in {install_mode} mode')

        # define if we need to link
        link = bool(install_mode == 'editable')

        # Determine the actual source path
        if link:
            actual_source = self._validate_component_path(component, custom_path)
        else:
            # Use the default source path
            actual_source = source_path
    
        # Install config directories
        os.makedirs(self.configpath, exist_ok=True)
        for directory in config_dirs:
            source = os.path.join(actual_source, directory)
            target = os.path.join(self.configpath, directory)
        
            self._copy_update_folder_file(source, target, link=link)
    
        # do the same for the templates
        template_source_base = os.path.join(actual_source, '..', 'templates')

        os.makedirs(self.templatepath, exist_ok=True)
        for directory in template_dirs:
            source = os.path.join(template_source_base, directory)
            target = os.path.join(self.templatepath, directory)
    
            if os.path.exists(source):
                self._copy_update_folder_file(source, target, link=link)

    def _determine_component_mode(self, core_arg, diag_arg, component):
        """
        Determine installation mode for a component based on CLI arguments.
        
        Args:
            core_arg: CLI argument for --core
            diag_arg: CLI argument for --diagnostics
            component (str): 'core' or 'diagnostics' - which component to check
            
        Returns:
            dict or False: Installation mode dict {'mode': str, 'path': str|None} 
                          or False to skip component
        """
        if component == 'core':
            if core_arg is not None:
                return self._get_install_mode(core_arg)
            if diag_arg is not None:
                return False
        else:  # diagnostics
            if diag_arg is not None:
                return self._get_install_mode(diag_arg)
            if core_arg is not None:
                return False
  
        # Neither specified: full install in standard mode
        return {'mode': 'standard', 'path': None}

    def install(self, args):
        """Install AQUA, find the folders and then install

        Args:
            args (argparse.Namespace): arguments from the command line
        """
        self.logger.info('Running the AQUA install')

        # configure where to install AQUA
        if args.path is None:
            self._config_home()
        else:
            self._config_path(args.path)

        # define the template path
        self.templatepath = os.path.join(self.configpath, 'templates')
    
        # Check current installation status
        install_info = self._check(silent=True, return_info=True)
        
        # Determine installation mode for each component
        core_mode = self._determine_component_mode(args.core, args.diagnostics, 'core')
        diag_mode = self._determine_component_mode(args.core, args.diagnostics, 'diagnostics')
        self.logger.debug('Installation modes - core: %s, diagnostics: %s', core_mode, diag_mode)

        if install_info['installed']:
            self.logger.warning('AQUA installation found in %s', self.configpath)
            if core_mode:
                self.logger.warning('Proceeding will remove the existing installation and all catalogs.')
                self.logger.warning('It will remove the diagnostics component if already installed.')
                if not self._remove_installation(confirm=True):
                    sys.exit()
            
            if core_mode is False and diag_mode:
                if not install_info['core']['installed']:
                    self.logger.error('Cannot install diagnostics without core. Install core first or use full installation.')
                    sys.exit(1)
                if install_info['diagnostics']['installed']:
                    self.logger.error('Diagnostics component is already installed. We cannot add it again, please use \'aqua uninstall\' before')
                    sys.exit(1)
                self.logger.info('Core already installed (%s mode), adding diagnostics component',
                           install_info['core']['mode'])
    
        # Validation: if installing only diagnostics, core must already be installed
        if core_mode is False and diag_mode:
            if not install_info['core']['installed']:
                self.logger.error('Cannot install diagnostics without core. Install core first or use full installation.')
                sys.exit(1)
            self.logger.info('Core already installed (%s mode), adding diagnostics component',
                           install_info['core']['mode'])

        # Install core
        if core_mode:
            self.logger.debug('Installing core component')
            self._install_component(
                component='core',
                mode=core_mode,
                source_path=self.corepath,
                config_dirs=CORE_CONFIG_DIRECTORIES,
                template_dirs=CORE_TEMPLATE_DIRECTORIES
            )
        
            self.logger.debug('Creating config-aqua.yaml configuration file')
            self._copy_update_folder_file(
                os.path.join(self.corepath, 'config-aqua.tmpl'),
                os.path.join(self.configpath, 'config-aqua.yaml')
            )

        # Install diagnostics if available and not skipped
        if diag_mode:
            if self.diagpath is not None:
                self.logger.debug('Installing diagnostics component')
                self._install_component(
                    component='diagnostics',
                    mode=diag_mode,
                    source_path=self.diagpath,
                    config_dirs=DIAGNOSTIC_CONFIG_DIRECTORIES,
                    template_dirs=DIAGNOSTIC_TEMPLATE_DIRECTORIES
                )
            else:
                if core_mode:
                    self.logger.warning('aqua.diagnostics package not found. Skipping diagnostics installation.')
                else:
                    self.logger.error('aqua.diagnostics package not found. Skipping diagnostics installation.')
                    sys.exit(1)

        # Create catalogs directory
        os.makedirs(os.path.join(self.configpath, CATPATH), exist_ok=True)

        # set machine
        self._set_machine(args)

    def _config_home(self):
        """Configure the AQUA installation folder, by default inside $HOME"""

        if 'HOME' in os.environ:
            path = os.path.join(os.environ['HOME'], '.aqua')
            self.configpath = path
            if not os.path.exists(path):
                os.makedirs(path, exist_ok=True)
            else:
                self.logger.debug('AQUA installation found in %s', path)
        else:
            self.logger.error('$HOME not found.'
                              'Please specify a path where to install AQUA and define AQUA_CONFIG as environment variable')
            sys.exit(1)

    def _config_path(self, path):
        """Define the AQUA installation folder when a path is specified

        Args:
            path (str): the path where to install AQUA
        """

        self.configpath = path
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
        else:
            if not os.path.isdir(path):
                self.logger.error("Path chosen is not a directory")
                sys.exit(1)

        check = query_yes_no(f"Do you want to create a link in the $HOME/.aqua to {path}", "yes")
        if check:
            if 'HOME' in os.environ:
                link = os.path.join(os.environ['HOME'], '.aqua')
                if os.path.exists(link) or os.path.islink(link):
                    self.logger.warning('Removing the content of %s', link)
                    if os.path.islink(link):
                        os.unlink(link)
                    elif os.path.isdir(link):
                        shutil.rmtree(link)
                    else:
                        os.remove(link)
                os.symlink(path, link)
            else:
                self.logger.error('$HOME not found. Cannot create a link to the installation path')
                self.logger.warning('AQUA will be installed in %s, but please remember to define AQUA_CONFIG environment variable', path)
        else:
            self.logger.warning('AQUA will be installed in %s, but please remember to define AQUA_CONFIG environment variable',
                                path)

    def _set_machine(self, args):
        """Modify the config-aqua.yaml with the identified machine"""

        if args.machine is not None:
            machine = args.machine
        else:
            machine = ConfigPath(configdir=self.configpath).get_machine()

        if machine is None:
            self.logger.info('Unknown machine!')
        else:
            self.configfile = os.path.join(self.configpath, 'config-aqua.yaml')
            self.logger.info('Setting machine name to %s', machine)

            with SafeFileLock(self.configfile + '.lock', loglevel=self.loglevel):
                cfg = load_yaml(self.configfile)
                cfg['machine'] = machine
                dump_yaml(self.configfile, cfg)

    def _remove_installation(self, confirm=True):
        """
        Remove AQUA installation directory
        
        Args:
            confirm (bool): Whether to ask for confirmation
            
        Returns:
            bool: True if removed, False if user cancelled
        """
        if confirm:
            check = query_yes_no(
                f"Do you want to remove AQUA installation in {self.configpath}. "
                "You will lose all catalogs installed.", "no"
            )
            if not check:
                return False

        self.logger.warning('Removing the content of %s', self.configpath)

        if os.path.islink(self.configpath):
            self.logger.info('Removing the link %s', self.configpath)
            os.unlink(self.configpath)
        else:
            self.logger.info('Removing directory %s', self.configpath)
            shutil.rmtree(self.configpath)

        return True

    def uninstall(self, args):
        """Remove AQUA"""
        print('Remove the AQUA installation')
        self._check()
        if not self._remove_installation(confirm=True):
            sys.exit()

    def update(self, args):
        """Update an existing catalog by copying it if not installed in editable mode"""

        self._check()
        if args.catalog:
            if args.catalog == 'all':
                self.logger.info('Updating all AQUA catalogs')
                catalogs = self._list_folder(f'{self.configpath}/{CATPATH}', return_list=True, silent=True)
                for catalog in catalogs:
                    print(f'Updating catalog {catalog} ..')
                    self._update_catalog(os.path.basename(catalog))
            else:
                self._update_catalog(args.catalog)
        else:
            self.logger.info('Updating AQUA installation...')

            # Update core if not in editable mode
            self._update_component(
                source_path=self.corepath,
                config_dirs=CORE_CONFIG_DIRECTORIES,
                template_dirs=CORE_TEMPLATE_DIRECTORIES,
                component_name='core'
            )

            # Update diagnostics if available and not in editable mode
            if self.diagpath is not None:
                self._update_component(
                    source_path=self.diagpath,
                    config_dirs=DIAGNOSTIC_CONFIG_DIRECTORIES,
                    template_dirs=DIAGNOSTIC_TEMPLATE_DIRECTORIES,
                    component_name='diagnostics'
                )

    def _update_component(self, source_path, config_dirs, template_dirs, component_name):
        """
        Update a component (core or diagnostics), skipping directories in editable mode.
        
        Args:
            source_path (str): Source path for the component
            config_dirs (list): List of config directories
            template_dirs (list): List of template directories
            component_name (str): Name of the component for logging
        """
        for directory in config_dirs:
            target = os.path.join(self.configpath, directory)
            
            # Skip if in editable mode
            if os.path.islink(target):
                self.logger.info(f'Skipping {component_name}/{directory} (editable mode)')
                continue
                
            source = os.path.join(source_path, directory)
            self._copy_update_folder_file(source, target, update=True)

        for directory in template_dirs:
            target = os.path.join(self.templatepath, directory)
            
            # Skip if in editable mode
            if os.path.islink(target):
                self.logger.info(f'Skipping {component_name}/{directory} template (editable mode)')
                continue
                
            source = os.path.join(source_path, '..', 'templates', directory)
            self._copy_update_folder_file(source, target, update=True)

    def _copy_update_folder_file(self, source, target, link=False, update=False):
        """Generic function to copy or update a source to a target folder"""

        # Check if the target exists
        if os.path.exists(target):
            if os.path.islink(target):
                self.logger.error('AQUA has been installed in editable mode, no need to update')
                sys.exit(1)
            # Update case
            if update:
                self.logger.info('Updating %s ...', target)
                if os.path.isdir(target):
                    shutil.rmtree(target)
                else:
                    os.remove(target)

        if os.path.exists(target):
            self.logger.error('%s already exist, please consider update or uninstall', target)
            sys.exit(1)

        # Handle linking
        if link:
            self.logger.info('Linking from %s to %s', source, target)
            os.symlink(source, target)

        # Handle copying
        else:
            if os.path.isdir(source):
                self.logger.info('Copying directory from %s to %s', source, target)
                shutil.copytree(source, target)
            else:
                self.logger.info('Copying file from %s to %s', source, target)
                shutil.copy2(source, target)

    def list(self, args):
        """List installed catalogs"""

        self._check()

        cdir = f'{self.configpath}/{CATPATH}'

        print('AQUA current installed catalogs in', cdir, ':')
        self._list_folder(cdir)

        if args.all:
            for content in CORE_CONFIG_DIRECTORIES + DIAGNOSTIC_CONFIG_DIRECTORIES:
                print(f'AQUA current installed {content} in {self.configpath}:')
                self._list_folder(os.path.join(self.configpath, content))

    @staticmethod
    def _list_folder(mydir, return_list=False, silent=False):
        """
        List all the files in a AQUA config folder and check if they are link or file/folder
        
        Args:
            mydir (str): the directory to be listed
            return_list (bool): if True, return the list of files for further processing
            silent (bool): if True, do not print the files, just return the list
        
        Returns:
            None or list: a list of files if return_list is True, otherwise nothing
        """

        list_files = []
        yaml_files = os.listdir(mydir)
        for file in yaml_files:
            file = os.path.join(mydir, file)
            if os.path.islink(file):
                orig_path = os.readlink(file)
                if not silent:
                    print(f"\t - {file} (editable from {orig_path})")
            else:
                if not silent:
                    print(f"\t - {file}")
                list_files.append(file)

        if return_list:
            return list_files
        
    @staticmethod
    def _get_config_dirs(component):
        """
        Get the config directories for a component
        
        Args:
            component (str): 'core' or 'diagnostics'

        Returns:
            list: List of config directories
        """
        if component == 'core':
            return CORE_CONFIG_DIRECTORIES
        if component == 'diagnostics':
            return DIAGNOSTIC_CONFIG_DIRECTORIES
        raise ValueError(f"Unknown component: {component}")
    
    @staticmethod
    def _get_install_mode(specific_arg):
        """
        Determine the installation mode for a component.
        
        Args:
            specific_arg: Value of --core or --diagnostics argument
            
        Returns:
            dict: {'mode': 'editable'|'standard', 'path': str|None}
        """
        # If specific_arg is a string path (not 'standard'), use editable mode
        if isinstance(specific_arg, str) and specific_arg != 'standard':
            return {'mode': 'editable', 'path': specific_arg}

        # If specific_arg is 'standard' (flag without value), use standard mode
        return {'mode': 'standard', 'path': None}



"""Tests for string unit conversion utilities"""

import pytest
import string
from aqua.core.util.string import unit_to_latex, generate_random_string
from aqua.core.util.string import get_quarter_anchor_month, clean_filename

@pytest.mark.aqua
@pytest.mark.parametrize("input_str, expected", [
    # Basic cases
    ("kg", "kg"),
    
    # Simple exponents (explicit and implicit)
    ("m^2", "m$^{2}$"),
    ("m**2", "m$^{2}$"),
    ("m2", "m$^{2}$"),
    ("m-2", "m$^{-2}$"),
    ("m^-2", "m$^{-2}$"),
    ("m**-2", "m$^{-2}$"),
    
    # Division notation
    ("W/m^2", "W m$^{-2}$"),
    ("m/s", "m s$^{-1}$"),
    ("kg/m/s", "kg m$^{-1}$ s$^{-1}$"),
    ("J/kg/K", "J kg$^{-1}$ K$^{-1}$"),
    
    # Grouped division
    ("W/(m^2 s)", "W m$^{-2}$ s$^{-1}$"),
    ("kg/(m s^2)", "kg m$^{-1}$ s$^{-2}$"),
    
    # Mixed notation
    ("kg m-1 s-1", "kg m$^{-1}$ s$^{-1}$"),
    ("kg m^-1 s^-1", "kg m$^{-1}$ s$^{-1}$"),
    ("kg m**-1 s**-1", "kg m$^{-1}$ s$^{-1}$"),
    
    # Special characters
    ("°C", "°C"),
    ("µg m^-3", "µg m$^{-3}$"),
    ("µg/m^3", "µg m$^{-3}$"),
    ("%", r"$\%$"),
    
    # Dimensionless units
    ("1", "1"),
    ("1 ", "1"),
    (" 1", "1"),

    # Edge cases
    ("", ""),
    (None, None),
    ("   ", ""),
    ("m/s/s", "m s$^{-1}$ s$^{-1}$"),
    
    # Already LaTeX (should be preserved)
    ("$\\mathrm{km}^2$", "$\\mathrm{km}^2$"),
    ("10^6 $\\mathrm{km}^2$", "10^6 $\\mathrm{km}^2$"),
    (r"$\mathrm{W} \mathrm{m}^{-2}$", r"$\mathrm{W} \mathrm{m}^{-2}$"),
    ("m^{2}", "m^{2}"), # Partial LaTeX
])
def test_unit_to_latex(input_str, expected):
    """Test unit_to_latex with various input formats"""
    assert unit_to_latex(input_str) == expected

@pytest.mark.aqua
def test_unit_to_latex_complex():
    """Test more complex combinations"""
    # Test complex mixed format
    assert unit_to_latex("W m-2 K-1") == "W m$^{-2}$ K$^{-1}$"
    # Test complex division with implicit exponents
    assert unit_to_latex("kg/m3") == "kg m$^{-3}$"
    # Test with extra spaces
    assert unit_to_latex(" W /  m^2 ") == "W m$^{-2}$"


@pytest.mark.aqua
def test_generate_random_string():
    """Test generate_random_string function"""
    # Test different lengths
    for length in [0, 1, 5, 10, 20]:
        result = generate_random_string(length)
        assert len(result) == length
        # Check that all characters are letters or digits
        assert all(c in string.ascii_letters + string.digits for c in result)


@pytest.mark.aqua
@pytest.mark.parametrize("freq_string, expected", [
    ("QE-DEC", "DEC"),  # Extract month when dash present
    ("Q-MAR", "MAR"),   # Different month
    ("QS", "DEC"),      # Default when no dash
])
def test_get_quarter_anchor_month(freq_string, expected):
    """Test get_quarter_anchor_month function"""
    assert get_quarter_anchor_month(freq_string) == expected


@pytest.mark.aqua
@pytest.mark.parametrize("filename, expected", [
    ("My File Name", "my_file_name"),  # Mixed case with spaces
    ("UPPER CASE", "upper_case"),      # Uppercase with spaces
    ("file_with_spaces.txt", "file_with_spaces.txt"), # Already clean
    ("  leading spaces  ", "__leading_spaces__"),     # Edge case: leading/trailing spaces
    ("", ""),
])
def test_clean_filename(filename, expected):
    """Test clean_filename function"""
    assert clean_filename(filename) == expected



"""Module for tests for AQUA cli"""

import os
import shutil
import sys
import subprocess
import pytest
from aqua.core.console.main import AquaConsole
from aqua.core.console.util import query_yes_no
from aqua.core.util import dump_yaml, load_yaml
from aqua import __version__ as version
from aqua import __path__ as pypath

TESTFILE = 'testfile.txt'
MACHINE = 'github'

pytestmark = [
    pytest.mark.aqua,
    pytest.mark.console
]

def set_args(args):
    """Helper function to simulate command line arguments"""
    sys.argv = ['aqua'] + args


@pytest.fixture(scope="session")
def tmpdir(tmp_path_factory):
    """Fixture to create a temporary directory"""
    mydir = tmp_path_factory.mktemp('tmp')
    yield mydir
    shutil.rmtree(str(mydir))


@pytest.fixture(scope="class")
def set_home():
    """Fixture to modify the HOME environment variable"""
    original_value = os.environ.get('HOME')

    def _modify_home(new_value):
        os.environ['HOME'] = new_value
    yield _modify_home
    os.environ['HOME'] = original_value


@pytest.fixture
def delete_home():
    """Fixture to delete the temporary HOME environment variable"""
    original_value = os.environ.get('HOME')

    def _modify_home():
        del os.environ['HOME']
    yield _modify_home
    os.environ['HOME'] = original_value


@pytest.fixture(scope="class")
def run_aqua_console_with_input(tmpdir):
    """Fixture to run AQUA console with some interactive command

    Args:
        tmpdir (str): temporary directory
    """
    def _run_aqua_console(args, input_text):
        """Run AQUA console with some interactive command

        Args:
            args (list): list of arguments
            input_text (str): input text
        """
        set_args(args)
        myfile = os.path.join(str(tmpdir), TESTFILE)
        with open(myfile, 'w', encoding='utf-8') as f:
            f.write(input_text)
        sys.stdin = open(myfile, 'r', encoding='utf-8')
        aquacli = AquaConsole()
        aquacli.execute()
        sys.stdin.close()
        os.remove(myfile)
    return _run_aqua_console


@pytest.fixture(scope="class")
def run_aqua():
    """Fixture to run AQUA console with some interactive command"""
    def _run_aqua_console(args):
        set_args(args)
        aquacli = AquaConsole()
        aquacli.execute()
    return _run_aqua_console


@pytest.fixture(scope="class")
def shared_aqua_install(tmpdir, set_home, run_aqua, run_aqua_console_with_input):
    """Shared AQUA installation for multiple tests in a class
    
    This fixture installs AQUA once and provides cleanup after all tests in the class.
    """
    mydir = str(tmpdir)
    set_home(mydir)
    
    # Install AQUA once for all tests in class
    run_aqua(['install', MACHINE])
    
    yield mydir
    
    # Cleanup after all tests in class
    if os.path.exists(os.path.join(mydir, '.aqua')):
        run_aqua_console_with_input(['uninstall'], 'yes')


@pytest.mark.aqua
class TestAquaConsole():
    """Class for AQUA console tests"""

    def test_console_install(self):
        """Test for CLI call"""
        # test version
        result = subprocess.run(['aqua', '--version'], check=False, capture_output=True, text=True)
        assert result.stdout.strip() == f'aqua v{version}'

        # test path
        result = subprocess.run(['aqua', '--path'], check=False, capture_output=True, text=True)
        assert pypath[0] == result.stdout.strip()

    # base set of tests
    def test_console_base(self, tmpdir, set_home, run_aqua, run_aqua_console_with_input):
        """Basic tests

        Args:
            tmpdir (str): temporary directory
            set_home (fixture): fixture to modify the HOME environment variable
            run_aqua (fixture): fixture to run AQUA console with some interactive command
            run_aqua_console_with_input (fixture): fixture to run AQUA console with some interactive command
        """

        # getting fixture
        mydir = str(tmpdir)
        set_home(mydir)

        # aqua install
        run_aqua(['install', MACHINE])
        assert os.path.isdir(os.path.join(mydir, '.aqua'))
        assert os.path.isfile(os.path.join(mydir, '.aqua', 'config-aqua.yaml'))

        # do it twice!
        run_aqua_console_with_input(['-vv', 'install', MACHINE], 'yes')
        assert os.path.exists(os.path.join(mydir, '.aqua'))
        for folder in ['fixes', 'data_model', 'grids']:
            assert os.path.isdir(os.path.join(mydir, '.aqua', folder))

        # add unexesting catalog from path
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['add', 'config/ueeeeee/ci'])
            assert excinfo.value.code == 1

        # add non existing catalog from default
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'add', 'antani'])
            assert excinfo.value.code == 1

        # add from wrongly formatted repository
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'add', 'pippo', '--repository', 'thisisnotauserandrepo'])
            assert excinfo.value.code == 1

        # add existing folder which is not a catalog
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['add', 'config/fixes'])
            assert excinfo.value.code == 1

        # create a test for DROP
        with pytest.raises(FileNotFoundError, match="ERROR: drop_config.yaml not found: you need to have this configuration file!"):
            run_aqua(['drop'])

        # create a test for catgen
        with pytest.raises(FileNotFoundError, match="ERROR: config.yaml not found: you need to have this configuration file!"):
            run_aqua(['catgen', '--config', 'config.yaml'])

        # uninstall and say no
        with pytest.raises(SystemExit) as excinfo:
            run_aqua_console_with_input(['uninstall'], 'no')
            assert excinfo.value.code == 0
            assert os.path.exists(os.path.join(mydir, '.aqua'))

        # uninstall and say yes
        run_aqua_console_with_input(['uninstall'], 'yes')
        assert not os.path.exists(os.path.join(mydir, '.aqua'))

    def test_console_drop(self, tmpdir, set_home, run_aqua, run_aqua_console_with_input): 
        """Test for running DROP via the console"""

        mydir = str(tmpdir)
        set_home(mydir)

        # aqua install
        run_aqua(['install', MACHINE])
        run_aqua(['add', 'ci', '--repository', 'DestinE-Climate-DT/Climate-DT-catalog'])

        # create fake config file
        drop_test = os.path.join(mydir, 'faketrip.yaml')
        dump_yaml(drop_test,{
                'target': {
                    'resolution': 'r200',
                    'frequency': 'monthly',
                    'catalog': 'ci'
                },
                'paths': {
                    'outdir': os.path.join(mydir, 'drop_test'),
                    'tmpdir': os.path.join(mydir, 'tmp')
                },
                'options': {
                    'loglevel': 'INFO'
                },
                'data': {
                    'IFS': {
                        'test-tco79': {
                            'long': {'vars': '2t'} 
                        }
                    }
                }
            }
        )

        # run DROP and verify that at least one file exist
        run_aqua(['drop', '--config', drop_test, '-w', '1', '-d', '--rebuild', '--startdate', '2020-01-01', '--enddate', '2020-03-31'])
        path = os.path.join(os.path.join(mydir, 'drop_test'),
                            "ci/IFS/test-tco79/r1/r200/monthly/mean/global/2t_ci_IFS_test-tco79_r1_r200_monthly_mean_global_202002.nc")
        assert os.path.isfile(path), f"File not found: {path}"

        # run DROP with a different stat and verify that at least one file exist
        run_aqua(['drop', '--config', drop_test, '-w', '1', '-d', '--rebuild', '--stat', 'min'])
        path = os.path.join(os.path.join(mydir, 'drop_test'),
                            "ci/IFS/test-tco79/r1/r200/monthly/min/global/2t_ci_IFS_test-tco79_r1_r200_monthly_min_global_202002.nc")
        assert os.path.isfile(path), f"File not found: {path}"
        
        # remove aqua
        run_aqua_console_with_input(['uninstall'], 'yes')

    @pytest.mark.parametrize("install_args,should_fail", [
        (['install', MACHINE, '--core'], False),
        (['install', MACHINE, '--diagnostics'], True),
    ])
    def test_console_selective_install(self, tmpdir, set_home, run_aqua, run_aqua_console_with_input, 
                                       install_args, should_fail):
        """Test for running selective install via the console (parametrized)"""

        mydir = str(tmpdir)
        set_home(mydir)

        if should_fail:
            # aqua install diagnostics only (should fail)
            with pytest.raises(SystemExit) as excinfo:
                run_aqua(install_args)
                assert excinfo.value.code == 1
        else:
            # aqua install core only
            run_aqua(install_args)
            assert os.path.isdir(os.path.join(mydir, '.aqua'))
            assert os.path.isfile(os.path.join(mydir, '.aqua', 'config-aqua.yaml'))

            # uninstall aqua
            run_aqua_console_with_input(['uninstall'], 'yes')
            assert not os.path.exists(os.path.join(mydir, '.aqua'))

    # def test_console_analysis(self, tmpdir, set_home, run_aqua, run_aqua_console_with_input):
    #     """Test for running the analysis via the console"""

    #     mydir = str(tmpdir)
    #     set_home(mydir)

    #     # aqua install
    #     run_aqua(['install', MACHINE])
    #     run_aqua(['add', 'ci', '--repository', 'DestinE-Climate-DT/Climate-DT-catalog'])

    #     test_dir = os.path.dirname(os.path.abspath(__file__))
    #     config_path = os.path.join(test_dir, 'analysis', 'config.aqua-analysis-test.yaml')

    #     # Run details
    #     catalog = 'ci'
    #     model = 'IFS'
    #     experiment = 'test-tco79'
    #     source = 'teleconnections'
    #     output_dir = os.path.join(mydir, 'output')
    #     regrid = False

    #     # run the analysis and verify that at least one file exist
    #     run_aqua(['analysis', '--config', config_path, '-m', model, '-e', experiment,
    #             '-s', source, '-d', output_dir, '-l', 'debug', '--regrid', regrid])
        
    #     output_path = os.path.join(output_dir, catalog, model, experiment, 'r1')
        
    #     assert os.path.exists(os.path.join(output_path, 'experiment.yaml')), \
    #         "experiment.yaml not found"
        
    #     log_file = os.path.join(output_path, 'dummy-dummy_tool.log')
    #     assert os.path.exists(log_file), \
    #         f"dummy-dummy_tool.log not found. Files in {output_path}: {os.listdir(output_path) if os.path.exists(output_path) else 'directory does not exist'}"
        
    #     # Check if "This is a dummy CLI script that does nothing." is in the log
    #     with open(log_file, 'r', encoding='utf-8') as f:
    #         content = f.read()
    #     assert "This is a dummy CLI script that does nothing." in content, \
    #         "Expected content not found in dummy-dummy_tool.log"
        
    #     assert os.path.exists(os.path.join(output_path, 'setup_checker.log')), \
    #         "setup_checker.log not found"

    #     # remove aqua
    #     run_aqua_console_with_input(['uninstall'], 'yes')


    def test_console_advanced(self, tmpdir, run_aqua, set_home, run_aqua_console_with_input):
        """Advanced tests for editable installation, editable catalog, catalog update,
        add a wrong catalog, uninstall

        Args:
            tmpdir (str): temporary directory
            run_aqua (fixture): fixture to run AQUA console with some interactive command
            set_home (fixture): fixture to modify the HOME environment variable
            run_aqua_console_with_input (fixture): fixture to run AQUA console with some interactive command
        """

        # getting fixture
        mydir = str(tmpdir)
        set_home(mydir)

        # check unexesting installation
        with pytest.raises(SystemExit) as excinfo:
            run_aqua_console_with_input(['uninstall'], 'yes')
            assert excinfo.value.code == 1

        # a new install
        run_aqua(['install', MACHINE])
        assert os.path.exists(os.path.join(mydir, '.aqua'))

        # add catalog again and error
        run_aqua(['-v', 'add', 'ci', '-e', 'AQUA_tests/catalog_copy'])
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'add', 'ci', '-e', 'config/catalogs/ci'])
            assert excinfo.value.code == 1
        assert os.path.exists(os.path.join(mydir, '.aqua/catalogs/ci'))

        # error for update an missing catalog
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'update', 'antani'])
            assert excinfo.value.code == 1

        # add non existing catalog editable
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'add', 'ci', '-e', 'config/catalogs/baciugo'])
            assert excinfo.value.code == 1
        assert not os.path.exists(os.path.join(mydir, '.aqua/catalogs/baciugo'))

        # remove existing catalog from link
        run_aqua(['remove', 'ci'])
        assert not os.path.exists(os.path.join(mydir, '.aqua/catalogs/ci'))


    def test_console_with_links(self, tmpdir, set_home, run_aqua_console_with_input):
        """Advanced tests for installation from path with symlinks"""

        # getting fixture
        mydir = str(tmpdir)
        set_home(mydir)

        # check unexesting installation
        with pytest.raises(SystemExit) as excinfo:
            run_aqua_console_with_input(['-v', 'install', MACHINE, '-p', 'environment.yml'], 'yes')
            assert excinfo.value.code == 1

        # install from path with grids
        # run_aqua_console_with_input(['-v', 'install', '-g', os.path.join(mydir, 'supercazzola')], 'yes')
        # assert os.path.exists(os.path.join(mydir, '.aqua'))

        # uninstall everything
        # run_aqua_console_with_input(['uninstall'], 'yes')
        # assert not os.path.exists(os.path.join(mydir,'.aqua'))

        # install from path
        run_aqua_console_with_input(['-v', 'install', MACHINE, '-p', os.path.join(mydir, 'vicesindaco')], 'yes')
        assert os.path.exists(os.path.join(mydir, 'vicesindaco'))

        # uninstall everything again
        run_aqua_console_with_input(['uninstall'], 'yes')
        assert not os.path.exists(os.path.join(mydir, '.aqua'))

    def test_console_editable(self, tmpdir, run_aqua, set_home, run_aqua_console_with_input):
        """Advanced tests for editable installation from path with editable mode"""

        # getting fixture
        mydir = str(tmpdir)
        set_home(mydir)

        # find the correct AQUA root and config paths
        test_dir = os.path.dirname(os.path.abspath(__file__))  # /path/to/AQUA/tests
        aqua_root = os.path.abspath(os.path.join(test_dir, '..'))  # /path/to/AQUA
        #config_dir = os.path.join(aqua_root, 'config')  # /path/to/AQUA/config

        # check unexesting installation
        #with pytest.raises(SystemExit) as excinfo:
        #    run_aqua(['-vv', 'install', MACHINE, '--core', test_dir])
        #    assert excinfo.value.code == 1

        # install from path with grids
        run_aqua(['-vv', 'install', MACHINE, '--core', aqua_root])
        assert os.path.exists(os.path.join(mydir, '.aqua'))
        for folder in ['fixes', 'data_model', 'grids']:
            assert os.path.islink(os.path.join(mydir, '.aqua', folder))
        assert os.path.isdir(os.path.join(mydir, '.aqua', 'catalogs'))

        # try to install diagnostics only on top of existing installation (should fail)
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['install', MACHINE, '--diagnostics', test_dir])
            assert excinfo.value.code == 1

        # install from path in editable mode
        run_aqua_console_with_input(['-vv', 'install', MACHINE, '--core', aqua_root,
                                     '--path', os.path.join(mydir, 'vicesindaco2')], 'yes')
        assert os.path.islink(os.path.join(mydir, '.aqua'))
        run_aqua_console_with_input(['uninstall'], 'yes')

        # install from path in editable mode but without aqua link
        run_aqua_console_with_input(['-vv', 'install', MACHINE, '--core', aqua_root,
                                     '--path', os.path.join(mydir, 'vicesindaco1')], 'no')
        assert not os.path.exists(os.path.join(mydir, '.aqua'))
        assert os.path.isdir(os.path.join(mydir, 'vicesindaco1', 'catalogs'))

        # uninstall everything again, using AQUA_CONFIG env variable
        os.environ['AQUA_CONFIG'] = os.path.join(mydir, 'vicesindaco1')
        run_aqua_console_with_input(['uninstall'], 'yes')
        assert not os.path.exists(os.path.join(mydir, 'vicesindaco1'))
        del os.environ['AQUA_CONFIG']

        assert not os.path.exists(os.path.join(mydir, '.aqua'))


    def test_console_without_home(self, delete_home, run_aqua, tmpdir, run_aqua_console_with_input):
        """Basic tests without HOME environment variable"""

        # getting fixture
        delete_home()
        mydir = str(tmpdir)

        print(f"HOME is set to: {os.environ.get('HOME')}")

        # check unexesting installation
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['install', MACHINE])
            assert excinfo.value.code == 1

        # install from path without home
        if os.path.exists(os.path.join(mydir, 'vicesindaco')):
            shutil.rmtree(os.path.join(mydir, 'vicesindaco'))
        run_aqua_console_with_input(['-v', 'install', MACHINE, '-p', os.path.join(mydir, 'vicesindaco')], 'yes')
        assert os.path.isdir(os.path.join(mydir, 'vicesindaco'))
        assert os.path.isfile(os.path.join(mydir, 'vicesindaco', 'config-aqua.yaml'))
        assert not os.path.exists(os.path.join(mydir, '.aqua'))


@pytest.mark.aqua
class TestAquaConsoleShared():
    """Tests that share a common AQUA installation to reduce I/O overhead"""

    def test_catalog_operations(self, shared_aqua_install, run_aqua):
        """Test catalog add, set, update, remove operations with shared installation"""
        mydir = shared_aqua_install

        # add two catalogs
        for catalog in ['ci', 'levante']:
            run_aqua(['add', catalog])
            assert os.path.isdir(os.path.join(mydir, '.aqua/catalogs', catalog))
            config_file = load_yaml(os.path.join(mydir, '.aqua', 'config-aqua.yaml'))
            assert catalog in config_file['catalog']

        # set catalog
        run_aqua(['set', 'ci'])
        assert os.path.isdir(os.path.join(mydir, '.aqua/catalogs/ci'))
        config_file = load_yaml(os.path.join(mydir, '.aqua', 'config-aqua.yaml'))
        assert config_file['catalog'][0] == 'ci'

        # update the installation files
        run_aqua(['-v', 'update'])
        assert os.path.isdir(os.path.join(mydir, '.aqua/fixes'))

        # update a catalog
        run_aqua(['-v', 'update', '-c', 'ci'])
        assert os.path.isdir(os.path.join(mydir, '.aqua/catalogs/ci'))

        # remove catalog
        run_aqua(['remove', 'ci'])
        assert not os.path.exists(os.path.join(mydir, '.aqua/catalogs/ci'))

    def test_editable_catalog_operations(self, shared_aqua_install, run_aqua):
        """Test editable catalog operations with shared installation"""
        mydir = shared_aqua_install

        # add catalog with editable option
        run_aqua(['-v', 'add', 'ci', '-e', 'AQUA_tests/catalog_copy'])
        assert os.path.isdir(os.path.join(mydir, '.aqua/catalogs/ci'))

        # update a catalog installed in editable mode (should fail)
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'update', '-c', 'ci'])
            assert excinfo.value.code == 1

        # error for update an editable catalog
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'update', 'ci'])
            assert excinfo.value.code == 1

        # remove existing catalog from link
        run_aqua(['remove', 'ci'])
        assert not os.path.exists(os.path.join(mydir, '.aqua/catalogs/ci'))

    def test_grids_and_fixes_operations(self, shared_aqua_install, run_aqua):
        """Test grids and fixes operations with shared installation"""
        mydir = shared_aqua_install

        # add mock grid file
        gridtest = os.path.join(mydir, 'supercazzola.yaml')
        dump_yaml(gridtest, {'grids': {'sindaco': {'path': '{{ grids }}/comesefosseantani.nc'}}})
        run_aqua(['-v', 'grids', 'add', gridtest])
        assert os.path.isfile(os.path.join(mydir, '.aqua/grids/supercazzola.yaml'))

        # add mock grid file but editable
        gridtest = os.path.join(mydir, 'garelli.yaml')
        dump_yaml(gridtest, {'grids': {'sindaco': {'path': '{{ grids }}/comesefosseantani.nc'}}})
        run_aqua(['-v', 'grids', 'add', gridtest, '-e'])
        assert os.path.islink(os.path.join(mydir, '.aqua/grids/garelli.yaml'))

        # remove grid file
        run_aqua(['-v', 'grids', 'remove', 'garelli.yaml'])
        assert not os.path.exists(os.path.join(mydir, '.aqua/grids/garelli.yaml'))

        # set the grids path in the config-aqua.yaml
        run_aqua(['-v', 'grids', 'set', os.path.join(mydir, 'pippo')])
        assert os.path.exists(os.path.join(mydir, 'pippo', 'grids'))
        assert os.path.exists(os.path.join(mydir, 'pippo', 'areas'))
        assert os.path.exists(os.path.join(mydir, 'pippo', 'weights'))
        config_file = load_yaml(os.path.join(mydir, '.aqua', 'config-aqua.yaml'))
        assert config_file['paths'] == {
            'grids': os.path.join(mydir, 'pippo', 'grids'),
            'areas': os.path.join(mydir, 'pippo', 'areas'),
            'weights': os.path.join(mydir, 'pippo', 'weights')
        }

        # add wrong fix file
        fixtest = os.path.join(mydir, 'antani.yaml')
        dump_yaml(fixtest, {'fixer_name':  'antani'})
        run_aqua(['fixes', 'add', fixtest])
        assert not os.path.exists(os.path.join(mydir, '.aqua/fixes/antani.yaml'))

        # error for already existing file
        gridtest = os.path.join(mydir, 'garelli.yaml')
        dump_yaml(gridtest, {'grids': {'sindaco': {'path': '{{ grids }}/comesefosseantani.nc'}}})
        run_aqua(['-v', 'grids', 'add', gridtest, '-e'])
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'grids', 'add', gridtest, '-e'])
            assert excinfo.value.code == 1

        # remove non existing grid file
        run_aqua(['-v', 'grids', 'remove', 'garelli.yaml'])
        assert not os.path.exists(os.path.join(mydir, '.aqua/grids/garelli.yaml'))

        # error for already non existing file
        with pytest.raises(SystemExit) as excinfo:
            run_aqua(['-v', 'fixes', 'remove', 'ciccio.yaml'])
            assert excinfo.value.code == 1

        # set the grids path in the config-aqua.yaml with the block already existing
        run_aqua(['-v', 'grids', 'set', os.path.join(mydir, 'pippo')])
        run_aqua(['-v', 'grids', 'set', os.path.join(mydir, 'pluto')])
        assert os.path.exists(os.path.join(mydir, 'pluto', 'grids'))
        config_file = load_yaml(os.path.join(mydir, '.aqua', 'config-aqua.yaml'))
        assert config_file['paths'] == {
            'grids': os.path.join(mydir, 'pluto', 'grids'),
            'areas': os.path.join(mydir, 'pluto', 'areas'),
            'weights': os.path.join(mydir, 'pluto', 'weights')
        }

        # base set of tests for list
    def test_console_list(self, shared_aqua_install, run_aqua, capfd):
        """Basic tests for list command"""

        # getting fixture
        mydir = shared_aqua_install

        run_aqua(['add', 'ci'])
        run_aqua(['add', 'ciccio', '-e', 'AQUA_tests/catalog_copy'])
        run_aqua(['list', '-a'])

        out, _ = capfd.readouterr()
        assert 'AQUA current installed catalogs in' in out
        assert 'ci' in out
        assert 'ciccio (editable' in out
        assert 'ifs.yaml' in out
        assert 'HealPix.yaml' in out

        run_aqua(['avail', '--repository', 'DestinE-Climate-DT/Climate-DT-catalog'])
        out, _ = capfd.readouterr()

        assert 'climatedt-phase1' in out
        assert 'lumi-phase1' in out

        run_aqua(['-v', 'update', '-c', 'all'])

        out, _ = capfd.readouterr()
        assert '.aqua/catalogs/ci ..' in out


class TestAquaConsoleGridBuilder():
    """Tests for the aqua grids build CLI."""

    @pytest.mark.parametrize(
        "command_args", [
            ['grids', 'build', '--model', 'ERA5', '--exp', 'era5-hpz3', '--source', 'monthly'],
        ]
    )
    def test_aqua_console_gridbuilder(self, run_aqua, command_args, tmpdir):
        """Test the aqua grids build CLI"""
        run_aqua(command_args + ['--verify', '--outdir', str(tmpdir)])

# checks for query function
@pytest.fixture
def run_query_with_input(tmpdir):
    def _run_query(input_text, default_answer):
        testfile = os.path.join(tmpdir, TESTFILE)
        with open(testfile, 'w', encoding='utf-8') as f:
            f.write(input_text)
        sys.stdin = open(testfile, 'r', encoding='utf-8')
        try:
            result = query_yes_no("Question?", default_answer)
        finally:
            sys.stdin.close()
            os.remove(testfile)
        return result
    return _run_query


@pytest.mark.aqua
class TestQueryYesNo:
    """Class for query_yes_no tests"""

    def test_query_yes_no_invalid_input(self, run_query_with_input):
        result = run_query_with_input("invalid\nyes", "yes")
        assert result is True

    def test_query_yes_no_explicit_yes(self, run_query_with_input):
        result = run_query_with_input("yes", "no")
        assert result is True

    def test_query_yes_no_explicit_no(self, run_query_with_input):
        result = run_query_with_input("no", "yes")
        assert result is False

    def test_query_yes_no_default(self, run_query_with_input):
        result = run_query_with_input("no", None)
        assert result is False



"""
Class for supplementary coordinate and dimension fixes beyond base data model.
Works BEFORE base data model transformation (DataModel/CoordTransformer).
"""
import xarray as xr
from aqua.core.logger import log_history, log_configure


class FixerDataModel:
    """
    Apply supplementary coordinate and dimension fixes beyond base data model.
    
    Works BEFORE base data model transformation by DataModel/CoordTransformer.
    Handles additional fixes specified in experiment-specific YAML files.
    
    This class handles:
    - Coordinate renaming (source → target)
    - Units override (without conversion)
    - Dimension renaming
    
    Args:
        fixes (dict): Dictionary containing the fixes from YAML file.
        loglevel (str): Log level for logging. Defaults to 'WARNING'.
        
    """

    def __init__(self, fixes=None, loglevel='WARNING'):
        self.fixes = fixes
        self.logger = log_configure(log_level=loglevel, log_name='FixerDataModel')
        self.loglevel = loglevel

    def apply(self, data: xr.Dataset) -> xr.Dataset:
        """
        Apply supplementary coordinate and dimension fixes.
        
        This method should be called AFTER DataModel.apply() has been run.
        It applies experiment-specific fixes that override or supplement
        the base data model transformations.

        Args:
            data (xr.Dataset): Dataset already processed by DataModel

        Returns:
            xr.Dataset: Dataset with supplementary fixes applied
        """
        if self.fixes is None:
            self.logger.debug("No supplementary fixes to apply")
            return data
        
        # Apply supplementary fixes from YAML
        data = self._fix_dims(data)
        data = self._fix_coord(data)

        return data

    def _fix_coord(self, data: xr.Dataset):
        """
        Apply supplementary coordinate fixes from fixes file.
        Handles coordinate renaming and units override.

        Args:
            data (xr.Dataset): Input dataset

        Returns:
            xr.Dataset: Dataset with fixed coordinates
        """
        coords_fix = self.fixes.get("coords", None)
        if not coords_fix:
            return data

        coords = list(coords_fix.keys())
        self.logger.debug("Supplementary coordinate fixes: %s", coords)

        for coord in coords:
            src_coord = coords_fix[coord].get("source", None)
            tgt_units = coords_fix[coord].get("tgt_units", None)

            if src_coord:
                if src_coord in data.coords:
                    data = data.rename({src_coord: coord})
                    self.logger.debug("Coordinate %s renamed to %s", src_coord, coord)
                    log_history(data[coord], f"Coordinate {src_coord} renamed to {coord} by FixerDataModel")
                else:
                    self.logger.warning("Coordinate %s not found", src_coord)

            if tgt_units:
                if coord in data.coords:
                    self.logger.debug("Coordinate %s units overridden to %s", coord, tgt_units)
                    self.logger.warning("Units override applied - no conversion performed")
                    data[coord].attrs['units'] = tgt_units
                    log_history(data[coord], f"Coordinate {coord} units set to {tgt_units} by FixerDataModel")
                else:
                    self.logger.warning("Coordinate %s not found", coord)

        return data
    
    def _fix_dims(self, data: xr.Dataset):
        """
        Apply supplementary dimension fixes from fixes file.

        Args:
            data (xr.Dataset): Input dataset

        Returns:
            xr.Dataset: Dataset with fixed dimensions
        """
        dims_fix = self.fixes.get("dims", None)
        if not dims_fix:
            return data

        dims = list(dims_fix.keys())
        self.logger.debug("Supplementary dimension fixes: %s", dims)

        for dim in dims:
            src_dim = dims_fix[dim].get("source", None)

            if src_dim and src_dim in data.dims:
                data = data.rename_dims({src_dim: dim})
                self.logger.debug("Dimension %s renamed to %s", src_dim, dim)
                log_history(data, f"Dimension {src_dim} renamed to {dim} by FixerDataModel")
            else:
                if src_dim:
                    self.logger.warning("Dimension %s not found", src_dim)

        return data



"""Common functions for the Reader"""
import xarray as xr


def check_catalog_source(cat, model, exp, source, name="dictionary"):
    """
    Check the entries of a nested dictionary based on the model/exp/source structure
    and return an updated source. The name argument can be used for proper printing.

    Args:
        cat (dict): The nested dictionary containing the catalog information.
        model (str): The model ID to check in the catalog.
        exp (str): The experiment ID to check in the catalog.
        source (str): The source ID to check in the catalog.
        name (str, optional): The name used for printing. Defaults to "dictionary".

    Returns:
        str: An updated source ID. If the source is not specified, "default"
            is chosen, or, if missing, the first source.
    """

    if model not in cat:
        avail = list(cat.keys())
        raise KeyError(f"Model {model} not found in {name}. " 
                       f"Please choose between available models: {avail}")
    if exp not in cat[model]:
        avail = list(cat[model].keys())
        raise KeyError(f"Experiment {exp} not found in {name} for model {model}. "
                       f"Please choose between available exps: {avail}")
    if not cat[model][exp].keys():
        raise KeyError(f"Experiment {exp} in {name} for model {model} has no sources.")

    if source:
        if source not in cat[model][exp]:
            if "default" not in cat[model][exp]:
                avail = list(cat[model][exp].keys())
                raise KeyError(f"Source {source} of experiment {exp} "
                               f"not found in {name} for model {model}. "
                               f"Please choose between available sources: {avail}")
            source = "default"
    else:
        source = list(cat[model][exp].keys())[0]  # take first source if none provided

    return source


def check_att(da, att):
    """
    Check if a dataarray has a specific attribute.

    Arguments:
        da (xarray.DataArray): DataArray to check
        att (dict or str): Attribute to check for

    Returns:
        Boolean
    """
    if att:
        if isinstance(att, str):
            return att in da.attrs
        elif isinstance(att, dict):
            key = list(att.keys())[0]
            if key in da.attrs:
                return da.attrs[key] == list(att.values())[0]
        else:
            return False
    else:
        return False


def set_attrs(ds, attrs):
    """
    Set an attribute for all variables in an xarray.Dataset

    Args:
        ds (xarray.Dataset or xarray.DataArray): Dataset to set attributes on
        attrs (dict): Dictionary of attributes to set
    
    Returns:
        xarray.Dataset or xarray.DataArray: Updated Dataset or DataArray, or the same object if not this.
    """
    if not isinstance(attrs, dict):
        raise TypeError("The 'attrs' argument must be a dictionary.")

    if isinstance(ds, xr.Dataset):
        for var in ds.data_vars:
            ds[var].attrs.update(attrs)
    elif isinstance(ds, xr.DataArray):
        ds.attrs.update(attrs)
    return ds




import os
import pytest

from aqua.core.configurer import ConfigPath
from aqua import show_catalog_content

@pytest.mark.aqua
def test_config_plain():
    config = ConfigPath()
    assert config.filename == 'config-aqua.yaml'
    assert 'ci' in config.catalog_available

@pytest.mark.aqua
def test_config_paths():
    configfile = 'tests/config/config-aqua-custom.yaml'

    configdir = ConfigPath().get_config_dir()

    # Copy the file to the config directory
    os.system(f'cp {configfile} {configdir}')

    config = ConfigPath(catalog='ci', filename='config-aqua-custom.yaml', configdir=configdir)
    paths, _ = config.get_machine_info()

    assert paths['paths']['grids'] == 'pluto'

    # Remove the copied file
    os.system(f'rm {configdir}/config-aqua-custom.yaml')

@pytest.mark.aqua
def test_show_catalog_content_basic():
    """Test show_catalog_content with no filters."""
    results = show_catalog_content()

    assert isinstance(results, dict)
    # Check structure: catalog -> model -> exp -> list of sources
    for _, catalog_data in results.items():
        assert isinstance(catalog_data, dict)
        for _, model_data in catalog_data.items():
            assert isinstance(model_data, dict)
            for _, sources in model_data.items():
                assert isinstance(sources, list)



'''
This module contains simple functions for hovmoller plots.
Their main purpose is to provide a simple interface for plotting
hovmoller diagrams with xarray DataArrays.
'''
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from aqua.core.util import create_folder, evaluate_colorbar_limits, unit_to_latex
from aqua.core.logger import log_configure
from .styles import ConfigStyle

# set default options for xarray
xr.set_options(keep_attrs=True)


def plot_hovmoller(data: xr.DataArray,
                   invert_axis=False,
                   invert_time=False,
                   invert_space_coord=False,
                   sym=False,
                   style=None,
                   contour=True,
                   dim='lon', figsize=(8, 13),
                   vmin=None, vmax=None, cmap='PuOr_r', # defaul map for MJO
                   title = None,
                   box_text=True,
                   cbar: bool = True,
                   text: list | str = None,
                   nlevels=8,
                   cbar_label=None,
                   cbar_orientation: str = 'horizontal',
                   return_fig=False,
                   fig: plt.Figure = None, ax: plt.Axes = None,
                   ax_pos: tuple = (1, 1, 1),
                   loglevel: str = "WARNING",
                   ):
    """
    Plot a hovmoller diagram for a given xarray DataArray.

    Args:
        data (xr.DataArray): The data to be plotted.
        invert_axis (bool): If True, the x-axis will be inverted.
        invert_time (bool): If True, the direction time axis will be inverted.
        invert_space_coord (bool): If True, the space coordinate axis will be inverted.
        sym (bool): If True, the color limits will be symmetric around zero.
        style (str): The style to be used for the plot. Default is None, which uses the default AQUA style.
        contour (bool): If True, contours will be plotted. If False, pcolormesh will be used.
        dim (str): The dimension to be averaged over. Default is 'lon'.
        figsize (tuple): Size of the figure. Default is (8, 13).
        vmin (float): Minimum value for the color limits. If None, it will be evaluated from the data.
        vmax (float): Maximum value for the color limits. If None, it will be evaluated from the data.
        cmap (str): Colormap to be used for the plot. Default is 'PuOr_r'.
        title (str): Title for the plot. If None, no title will be set.
        box_text (bool): If True, a box with the min and max values of the dimension will be added to the plot.
        cbar (bool): If True, a colorbar will be added to the plot.
        text (list | str): Text to be added to the plot. If None, no text will be added.
        nlevels (int): Number of contour levels. Default is 8.
        cbar_label (str): Label for the colorbar. If None, it will be generated from the data attributes.
        cbar_orientation (str): Orientation of the colorbar. Default is 'horizontal'.
        return_fig (bool): If True, the figure and axes will be returned. Default is False.
        fig (plt.Figure): Matplotlib figure object to plot on. If None, a new figure will be created.
        ax (plt.Axes): Matplotlib axes object to plot on. If None, a new axes will be created.
        ax_pos (tuple): Position of the axes in the figure. Default is (1, 1, 1), which means a single subplot.
        loglevel (str): Logging level. Default is "WARNING".

    Returns:
        plt.Figure: The matplotlib figure object containing the hovmoller plot.
        plt.Axes: The matplotlib axes object containing the hovmoller plot.
    """
    logger = log_configure(log_level=loglevel, log_name='Hovmoller')
    ConfigStyle(style=style, loglevel=loglevel)

    # Check if data is a DataArray
    if not isinstance(data, xr.DataArray):
        raise TypeError('Data is not a DataArray')

    if dim:
        # Evaluate the mean over the dimension to be averaged over
        logger.info('Averaging over dimension: {}'.format(dim))
        dim_min = data.coords[dim].min()
        dim_max = data.coords[dim].max()
        dim_min = np.round(dim_min, 0)
        dim_max = np.round(dim_max, 0)

        data_mean = data.mean(dim=dim)
    else:
        data_mean = data

    # Create figure and axes
    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot(ax_pos[0], ax_pos[1], ax_pos[2])

    # Plot the data
    if invert_axis:
        logger.debug('Inverting axis for plot')
        x = data_mean.coords[data_mean.dims[-1]]
        y = data_mean.coords['time']
        ax.set_xlabel(data_mean.dims[-1])
        ax.set_ylabel('time')
        if invert_time:
            plt.gca().invert_yaxis()
        if invert_space_coord:
            plt.gca().invert_xaxis()
    else:
        x = data_mean.coords['time']
        y = data_mean.coords[data_mean.dims[-1]]
        ax.set_xlabel('time')
        ax.set_ylabel(data_mean.dims[-1])
        if invert_time:
            plt.gca().invert_xaxis()
        if invert_space_coord:
            plt.gca().invert_yaxis()

    if vmin is None or vmax is None:
        vmin, vmax = evaluate_colorbar_limits(maps=[data], sym=sym)
    else:
        if sym:
            logger.warning("sym=True, vmin and vmax given will be ignored")
            vmin, vmax = evaluate_colorbar_limits(maps=[data], sym=sym)
    logger.debug("Setting vmin to %s, vmax to %s", vmin, vmax)

    if contour:
        try:
            levels = np.linspace(vmin, vmax, nlevels)
            if invert_axis:
                im = ax.contourf(x, y, data_mean, levels=levels, cmap=cmap,
                                 vmin=vmin, vmax=vmax, extend='both')
            else:
                im = ax.contourf(x, y, data_mean.T, levels=levels, cmap=cmap,
                                 vmin=vmin, vmax=vmax, extend='both')
        except TypeError as e:
            logger.warning('Could not evaluate levels: {}'.format(e))
            if invert_axis:
                im = ax.contourf(x, y, data_mean, cmap=cmap,
                                 vmin=vmin, vmax=vmax, extend='both')
            else:
                im = ax.contourf(x, y, data_mean.T, cmap=cmap,
                                 vmin=vmin, vmax=vmax, extend='both')
    else:  # pcolormesh
        if invert_axis:
            im = ax.pcolormesh(x, y, data_mean, cmap=cmap,
                               vmin=vmin, vmax=vmax)
        else:
            im = ax.pcolormesh(x, y, data_mean.T, cmap=cmap,
                               vmin=vmin, vmax=vmax)

    if cbar is True:
        # Adjust the location of the subplots on the page to make room for the colorbar
        fig.subplots_adjust(bottom=0.25, top=0.9, left=0.05, right=0.95,
                            wspace=0.1, hspace=0.5)
        box = ax.get_position()
        cbar_ax = fig.add_axes([
            box.x0 - 0.01 + box.width + 0.01,  # a small gap to the right
            box.y0,                     # same vertical start
            0.015,                      # narrow width
            box.height                  # same height
        ])
        # cbar label
        if cbar_label is None:
            try:
                var_name = data_mean.short_name
            except AttributeError:
                try:
                    var_name = data_mean.long_name
                except AttributeError:
                    var_name = None
            # Add units
            try:
                units = data_mean.units
            except AttributeError:
                units = None
            if var_name is not None and units is not None:
                cbar_label = '{} [{}]'.format(var_name, unit_to_latex(units))
            elif var_name is not None:
                cbar_label = var_name
            else:
                cbar_label = None
                logger.warning('Could not find a label for the colorbar')

        fig.colorbar(im, cax=cbar_ax, orientation=cbar_orientation,
                label=cbar_label)

    if box_text:
        # Add min and max values of the dim on the top right corner
        ax.text(0.99, 0.99, f'{dim} = {dim_min:.2f} to {dim_max:.2f}',
                verticalalignment='top', horizontalalignment='right',
                transform=ax.transAxes, fontsize=12,
                bbox=dict(facecolor='white', alpha=0.5))
    if text:
        logger.debug("Adding text in the plot: %s", text)
        ax.text(-0.3, 0.33, text, fontsize=15, color='dimgray', rotation=90, transform=ax.transAxes, ha='center')

    if title:    
        ax.set_title(title, fontsize=20)

    if return_fig:
        logger.debug("Returning figure and axes")
        return fig, ax
